# Standard library imports.
import sys

# Enthought library imports.
from pyface.tasks.i_editor_area_pane import IEditorAreaPane, \
    MEditorAreaPane
from traits.api import Bool, cached_property, Callable, Dict, Instance, List, \
    on_trait_change, Property, provides, Str
from pyface.qt import QtCore, QtGui
from pyface.action.api import Action, Group, MenuManager
from pyface.tasks.task_layout import PaneItem, Tabbed, Splitter
from pyface.mimedata import PyMimeData
from pyface.api import FileDialog
from pyface.constant import OK
from pyface.drop_handler import IDropHandler, BaseDropHandler, FileDropHandler

# Local imports.
from .task_pane import TaskPane

###############################################################################
# 'SplitEditorAreaPane' class.
###############################################################################

@provides(IEditorAreaPane)
class SplitEditorAreaPane(TaskPane, MEditorAreaPane):
    """ The toolkit-specific implementation of an SplitEditorAreaPane.

    See the IEditorAreaPane interface for API documentation.
    """


    #### SplitEditorAreaPane interface #####################################

    # Currently active tabwidget
    active_tabwidget = Instance(QtGui.QTabWidget)

    # list of installed drop handlers
    drop_handlers = List(IDropHandler)

    # Additional callback functions. Few useful callbacks that can be included:
    #  'new': new file action (takes no argument)
    #  'open': open file action (takes file_path as single argument)
    #  'open_dialog': show the open file dialog (responsibility of the callback,
    #     takes no argument), overrides 'open' callback
    # They are used to create shortcut buttons for these actions in the empty
    # pane that gets created when the user makes a split
    callbacks = Dict({}, key=Str, value=Callable)

    # The constructor of the empty widget which comes up when one creates a split
    create_empty_widget = Callable

    #### Private interface ###################################################

    _private_drop_handlers = List(IDropHandler)
    _all_drop_handlers = Property(
        List(IDropHandler),
        depends_on=['drop_handlers', '_private_drop_handlers']
    )

    def __private_drop_handlers_default(self):
        """ By default, two private drop handlers are installed:

            1. For dropping of tabs from one pane to other
            2. For dropping of supported files from file-browser pane or outside
            the application
        """
        return [TabDropHandler(),
                FileDropHandler(extensions=self.file_drop_extensions,
                                open_file=lambda path:self.trait_set(file_dropped=path))]

    @cached_property
    def _get__all_drop_handlers(self):
        return self.drop_handlers + self._private_drop_handlers

    def _create_empty_widget_default(self):
        return lambda : self.active_tabwidget.create_empty_widget()

    ###########################################################################
    # 'TaskPane' interface.
    ###########################################################################

    def create(self, parent):
        """ Create and set the toolkit-specific control that represents the
            pane.
        """
        # Create and configure the Editor Area Widget.
        self.control = EditorAreaWidget(self, parent)
        self.active_tabwidget = self.control.tabwidget()

        # handle application level focus changes
        QtGui.QApplication.instance().focusChanged.connect(self._focus_changed)

        # set key bindings
        self.set_key_bindings()

    def destroy(self):
        """ Destroy the toolkit-specific control that represents the pane.
        """
        # disconnect application level focus change signals first, else it gives
        # weird runtime errors trying to access non-existent objects
        QtGui.QApplication.instance().focusChanged.disconnect(self._focus_changed)

        for editor in self.editors[:]:
            self.remove_editor(editor)

        super(SplitEditorAreaPane, self).destroy()

    ###########################################################################
    # 'IEditorAreaPane' interface.
    ###########################################################################

    def activate_editor(self, editor):
        """ Activates the specified editor in the pane.
        """
        active_tabwidget = editor.control.parent().parent()
        active_tabwidget.setCurrentWidget(editor.control)
        self.active_tabwidget = active_tabwidget
        editor_widget = editor.control.parent()
        editor_widget.setVisible(True)
        editor_widget.raise_()
        # Set focus on last active widget in editor if possible
        if editor.control.focusWidget():
            editor.control.focusWidget().setFocus()
        else:
            editor.control.setFocus()
        # Set active_editor at the end of the method so that the notification
        # occurs when everything is ready.
        self.active_editor = editor

    def add_editor(self, editor):
        """ Adds an editor to the active_tabwidget
        """
        editor.editor_area = self
        editor.create(self.active_tabwidget)
        index = self.active_tabwidget.addTab(editor.control,
                                             self._get_label(editor))
        # There seem to be a bug in pyside or qt, where the index is set to 1
        # when you create the first tab. This is a hack to fix it.
        if self.active_tabwidget.count() == 1:
            index = 0
        self.active_tabwidget.setTabToolTip(index, editor.tooltip)
        self.editors.append(editor)

    def remove_editor(self, editor):
        """ Removes an editor from the associated tabwidget
        """
        tabwidget = editor.control.parent().parent()
        tabwidget.removeTab(tabwidget.indexOf(editor.control))
        self.editors.remove(editor)
        editor.destroy()
        editor.editor_area = None
        if not self.editors:
            self.active_editor = None


    ##########################################################################
    # 'IAdvancedEditorAreaPane' interface.
    ##########################################################################

    def get_layout(self):
        """ Returns a LayoutItem that reflects the current state of the
        tabwidgets in the split framework.
        """
        return self.control.get_layout()

    def set_layout(self, layout):
        """ Applies the given LayoutItem.
        """
        self.control.set_layout(layout)

    ##########################################################################
    # 'SplitEditorAreaPane' interface.
    ##########################################################################

    def get_context_menu(self, pos):
        """ Returns a context menu containing split/collapse actions

        pos : position (in global coordinates) where the context menu was
        requested
        """
        menu = MenuManager()
        splitter = None

        splitter = None
        for tabwidget in self.tabwidgets():
            # obtain tabwidget's bounding rectangle in global coordinates
            global_rect = QtCore.QRect(tabwidget.mapToGlobal(QtCore.QPoint(0, 0)),
                                        tabwidget.size())
            if global_rect.contains(pos):
                splitter = tabwidget.parent()

        # no split/collapse context menu for positions outside any tabwidget
        # region
        if not splitter:
            return

        # add split actions (only show for non-empty tabwidgets)
        if not splitter.is_empty():
            actions = [Action(id='split_hor', name='Create new pane to the right',
                       on_perform=lambda : splitter.split(orientation=
                        QtCore.Qt.Horizontal)),
                       Action(id='split_ver', name='Create new pane to the bottom',
                       on_perform=lambda : splitter.split(orientation=
                        QtCore.Qt.Vertical))]

            splitgroup = Group(*actions, id='split')
            menu.append(splitgroup)

        # add collapse action (only show for collapsible splitters)
        if splitter.is_collapsible():
            if splitter is splitter.parent().leftchild:
                if splitter.parent().orientation() is QtCore.Qt.Horizontal:
                    text = 'Merge with right pane'
                else:
                    text = 'Merge with bottom pane'
            else:
                if splitter.parent().orientation() is QtCore.Qt.Horizontal:
                    text = 'Merge with left pane'
                else:
                    text = 'Merge with top pane'
            actions = [Action(id='merge', name=text,
                        on_perform=lambda : splitter.collapse())]

            collapsegroup = Group(*actions, id='collapse')
            menu.append(collapsegroup)

        # return QMenu object
        return menu

    ###########################################################################
    # Protected interface.
    ###########################################################################

    def _get_label(self, editor):
        """ Return a tab label for an editor.
        """
        try:
            label = editor.name
            if editor.dirty:
                label = '*' + label
        except AttributeError:
            label = ''
        return label

    def _get_editor(self, editor_widget):
        """ Returns the editor corresponding to editor_widget
        """
        for editor in self.editors:
            if editor.control is editor_widget:
                return editor
        return None

    def set_key_bindings(self):
        """ Set keyboard shortcuts for tabbed navigation
        """
        # Add shortcuts for scrolling through tabs.
        if sys.platform == 'darwin':
            next_seq = 'Ctrl+}'
            prev_seq = 'Ctrl+{'
        else:
            next_seq = 'Ctrl+PgDown'
            prev_seq = 'Ctrl+PgUp'
        shortcut = QtGui.QShortcut(QtGui.QKeySequence(next_seq), self.control)
        shortcut.activated.connect(self._next_tab)
        shortcut = QtGui.QShortcut(QtGui.QKeySequence(prev_seq), self.control)
        shortcut.activated.connect(self._previous_tab)

        # Add shortcuts for switching to a specific tab.
        mod = 'Ctrl+' if sys.platform == 'darwin' else 'Alt+'
        mapper = QtCore.QSignalMapper(self.control)
        mapper.mapped.connect(self._activate_tab)
        for i in range(1, 10):
            sequence = QtGui.QKeySequence(mod + str(i))
            shortcut = QtGui.QShortcut(sequence, self.control)
            shortcut.activated.connect(mapper.map)
            mapper.setMapping(shortcut, i - 1)

    def _activate_tab(self, index):
        """ Activates the tab with the specified index, if there is one.
        """
        self.active_tabwidget.setCurrentIndex(index)
        current_widget = self.active_tabwidget.currentWidget()
        for editor in self.editors:
            if current_widget == editor.control:
                self.activate_editor(editor)

    def _next_tab(self):
        """ Activate the tab after the currently active tab.
        """
        index = self.active_tabwidget.currentIndex()
        new_index = index + 1 if index < self.active_tabwidget.count() - 1 else 0
        self._activate_tab(new_index)

    def _previous_tab(self):
        """ Activate the tab before the currently active tab.
        """
        index = self.active_tabwidget.currentIndex()
        new_index = index - 1 if index > 0  else self.active_tabwidget.count() - 1
        self._activate_tab(new_index)

    def tabwidgets(self):
        """ Returns the list of tabwidgets associated with the current editor
        area.
        """
        return self.control.tabwidgets()

    #### Trait change handlers ################################################

    @on_trait_change('editors:[dirty, name]')
    def _update_label(self, editor, name, new):
        index = self.active_tabwidget.indexOf(editor.control)
        self.active_tabwidget.setTabText(index, self._get_label(editor))

    @on_trait_change('editors:tooltip')
    def _update_tooltip(self, editor, name, new):
        index = self.active_tabwidget.indexOf(editor.control)
        self.active_tabwidget.setTabToolTip(index, self._get_label(editor))

    #### Signal handlers ######################################################

    def _find_ancestor_draggable_tab_widget(self, control):
        """ Find the draggable tab widget to which a widget belongs. """

        while not isinstance(control, DraggableTabWidget):
            control = control.parent()

        return control

    def _focus_changed(self, old, new):
        """Set the active tabwidget after an application-level change in focus.
        """

        if new is not None:
            if isinstance(new, DraggableTabWidget):
                if new.editor_area == self:
                    self.active_tabwidget = new
            elif isinstance(new, QtGui.QTabBar):
                if self.control.isAncestorOf(new):
                    self.active_tabwidget = \
                        self._find_ancestor_draggable_tab_widget(new)
            else:
                # Check if any of the editor widgets have focus.
                # If so, make it active.
                for editor in self.editors:
                    control = editor.control
                    if control is not None and control.isAncestorOf(new):
                        active_tabwidget = \
                            self._find_ancestor_draggable_tab_widget(control)
                        active_tabwidget.setCurrentWidget(control)
                        self.active_tabwidget = active_tabwidget
                        break

    def _active_tabwidget_changed(self, new):
        """Set the active editor whenever the active tabwidget updates.
        """

        if new is None or new.parent().is_empty():
            active_editor = None
        else:
            active_editor = self._get_editor(new.currentWidget())

        self.active_editor = active_editor


###############################################################################
# Auxiliary classes.
###############################################################################

class EditorAreaWidget(QtGui.QSplitter):
    """ Container widget to hold a QTabWidget which are separated by other
    QTabWidgets via splitters.

    An EditorAreaWidget is essentially a Node object in the editor area layout
    tree.
    """

    def __init__(self, editor_area, parent=None, tabwidget=None):
        """ Creates an EditorAreaWidget object.

        editor_area : global SplitEditorAreaPane instance
        parent : parent splitter
        tabwidget : tabwidget object contained by this splitter

        """
        super(EditorAreaWidget, self).__init__(parent=parent)
        self.editor_area = editor_area

        if not tabwidget:
            tabwidget = DraggableTabWidget(editor_area=self.editor_area,
                                        parent=self)

        # add the tabwidget to the splitter
        self.addWidget(tabwidget)
        # showing the tabwidget after reparenting
        tabwidget.show()

        # Initializes left and right children to None (since no initial splitter
        # children are present)
        self.leftchild = None
        self.rightchild = None

    def get_layout(self):
        """ Returns a LayoutItem that reflects the layout of the current
        splitter.
        """
        ORIENTATION_MAP = {QtCore.Qt.Horizontal: 'horizontal',
                           QtCore.Qt.Vertical: 'vertical'}
        # obtain layout based on children layouts
        if not self.is_leaf():
            layout = Splitter(self.leftchild.get_layout(),
                            self.rightchild.get_layout(),
                            orientation=ORIENTATION_MAP[self.orientation()])
        # obtain the Tabbed layout
        else:
            if self.is_empty():
                layout = Tabbed(PaneItem(id=-1,
                                        width=self.width(),
                                        height=self.height()),
                                active_tab=0)
            else:
                items = []
                for i in range(self.tabwidget().count()):
                    widget = self.tabwidget().widget(i)
                    # mark identification for empty_widget
                    editor = self.editor_area._get_editor(widget)
                    item_id = self.editor_area.editors.index(editor)
                    item_width = self.width()
                    item_height = self.height()
                    items.append(PaneItem(id=item_id,
                                        width=item_width,
                                        height=item_height))
                layout = Tabbed(*items, active_tab=self.tabwidget().currentIndex())
        return layout

    def set_layout(self, layout):
        """ Applies the given LayoutItem to current splitter.
        """
        ORIENTATION_MAP = {'horizontal': QtCore.Qt.Horizontal,
                           'vertical': QtCore.Qt.Vertical}
        # if not a leaf splitter
        if isinstance(layout, Splitter):
            self.split(orientation=ORIENTATION_MAP[layout.orientation])
            self.leftchild.set_layout(layout=layout.items[0])
            self.rightchild.set_layout(layout=layout.items[1])

            # setting sizes of children along splitter direction
            if layout.orientation=='horizontal':
                sizes = [self.leftchild.width(), self.rightchild.width()]
                self.resize(sum(sizes), self.leftchild.height())
            else:
                sizes = [self.leftchild.height(), self.rightchild.height()]
                self.resize(self.leftchild.width(), sum(sizes))
            self.setSizes(sizes)

        # if it is a leaf splitter
        elif isinstance(layout, Tabbed):
            # don't clear-out empty_widget's information if all it contains is an
            # empty_widget
            if not self.is_empty():
                self.tabwidget().clear()

            for item in layout.items:
                if not item.id==-1:
                    editor = self.editor_area.editors[item.id]
                    self.tabwidget().addTab(editor.control,
                                            self.editor_area._get_label(editor))
                self.resize(item.width, item.height)
            self.tabwidget().setCurrentIndex(layout.active_tab)

    def tabwidget(self):
        """ Obtain the tabwidget associated with current EditorAreaWidget
        (returns None for non-leaf splitters)
        """
        for child in self.children():
            if isinstance(child, QtGui.QTabWidget):
                return child
        return None

    def tabwidgets(self):
        """ Return a list of tabwidgets associated with current splitter or
        any of its descendants.
        """
        tabwidgets = []
        if self.is_leaf():
            tabwidgets.append(self.tabwidget())

        else:
            tabwidgets.extend(self.leftchild.tabwidgets())
            tabwidgets.extend(self.rightchild.tabwidgets())

        return tabwidgets

    def sibling(self):
        """ Returns another child of its parent. Returns None if it can't find
        any sibling.
        """
        parent = self.parent()

        if self.is_root():
            return None

        if self is parent.leftchild:
            return parent.rightchild
        elif self is parent.rightchild:
            return parent.leftchild

    def is_root(self):
        """ Returns True if the current EditorAreaWidget is the root widget.
        """
        parent = self.parent()
        if isinstance(parent, EditorAreaWidget):
            return False
        else:
            return True

    def is_leaf(self):
        """ Returns True if the current EditorAreaWidget is a leaf, i.e., it has
        a tabwidget as one of it's immediate child.
        """
        # a leaf has it's leftchild and rightchild None
        if not self.leftchild and not self.rightchild:
            return True
        return False

    def is_empty(self):
        """ Returns True if the current splitter's tabwidget doesn't contain any
        tab.
        """
        return bool(self.tabwidget().empty_widget)

    def is_collapsible(self):
        """ Returns True if the current splitter can be collapsed to its sibling,
        i.e. if it is (a) either empty, or (b) it has a sibling which is a leaf.
        """
        if self.is_root():
            return False

        if self.is_empty():
            return True

        sibling = self.sibling()

        if sibling.is_leaf():
            return True
        else:
            return False

    def split(self, orientation=QtCore.Qt.Horizontal):
        """ Split the current splitter into two children splitters. The current
        splitter's tabwidget is moved to the left child while a new empty
        tabwidget is added to the right child.

        orientation : whether to split horizontally or vertically
        """
        # set splitter orientation
        self.setOrientation(orientation)
        orig_size = self.sizes()[0]

        # create new children
        self.leftchild = EditorAreaWidget(self.editor_area, parent=self,
                                        tabwidget=self.tabwidget())
        self.rightchild = EditorAreaWidget(self.editor_area, parent=self,
                                        tabwidget=None)

        # add newly generated children
        self.addWidget(self.leftchild)
        self.addWidget(self.rightchild)

        # set equal sizes of splits
        self.setSizes([orig_size/2,orig_size/2])

        # make the rightchild's tabwidget active & show its empty widget
        self.editor_area.active_tabwidget = self.rightchild.tabwidget()

    def collapse(self):
        """ Collapses the current splitter and its sibling splitter to their
        parent splitter. Merges together the tabs of both's tabwidgets.

        Does nothing if the current splitter is not collapsible.
        """
        if not self.is_collapsible():
            return

        parent = self.parent()
        sibling = self.sibling()

        # this will happen only if self is empty, else it will not be
        # collapsible at all
        if sibling and (not sibling.is_leaf()):
            parent.setOrientation(sibling.orientation())
            # reparent sibling's children to parent
            parent.addWidget(sibling.leftchild)
            parent.addWidget(sibling.rightchild)
            parent.leftchild = sibling.leftchild
            parent.rightchild = sibling.rightchild
            # blindly make the first tabwidget active as it is not clear which
            # tabwidget should get focus now (FIXME??)
            self.editor_area.active_tabwidget = parent.tabwidgets()[0]
            self.setParent(None)
            sibling.setParent(None)
            return

        # save original currentwidget to make active later
        # (if self is empty, make the currentwidget of sibling active)
        if not self.is_empty():
            orig_currentWidget = self.tabwidget().currentWidget()
        else:
            orig_currentWidget = sibling.tabwidget().currentWidget()

        left = parent.leftchild.tabwidget()
        right = parent.rightchild.tabwidget()
        target = DraggableTabWidget(editor_area=self.editor_area, parent=parent)

        # add tabs of left and right tabwidgets to target
        for source in (left, right):
            # Note: addTab removes widgets from source tabwidget, so
            # grabbing all the source widgets beforehand
            # (not grabbing empty_widget)
            widgets = [source.widget(i) for i in range(source.count()) if not
                        source.widget(i) is source.empty_widget]
            for editor_widget in widgets:
                editor = self.editor_area._get_editor(editor_widget)
                target.addTab(editor_widget,
                            self.editor_area._get_label(editor))

        # add target to parent
        parent.addWidget(target)

        # make target the new active tabwidget and make the original focused
        # widget active in the target too
        self.editor_area.active_tabwidget = target
        target.setCurrentWidget(orig_currentWidget)

        # remove parent's splitter children
        parent.leftchild = None
        parent.rightchild = None
        self.setParent(None)
        sibling.setParent(None)


class DraggableTabWidget(QtGui.QTabWidget):
    """ Implements a QTabWidget with event filters for tab drag and drop
    """

    def __init__(self, editor_area, parent):
        """
        editor_area : global SplitEditorAreaPane instance
        parent : parent of the tabwidget
        """
        super(DraggableTabWidget, self).__init__(parent)
        self.editor_area = editor_area

        # configure QTabWidget
        self.setTabBar(DraggableTabBar(editor_area=editor_area, parent=self))
        self.setDocumentMode(True)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setFocusProxy(None)
        self.setMovable(False) # handling move events myself
        self.setTabsClosable(True)
        self.setAutoFillBackground(True)

        # set drop and context menu policies
        self.setAcceptDrops(True)
        self.setContextMenuPolicy(QtCore.Qt.DefaultContextMenu)

        # connecting signals
        self.tabCloseRequested.connect(self._close_requested)
        self.currentChanged.connect(self._current_changed)

        # shows the custom empty widget containing buttons for relevant actions
        self.show_empty_widget()

    def show_empty_widget(self):
        """ Shows the empty widget (containing buttons to open new file, and
        collapse the split).
        """
        self.empty_widget = None
        self.editor_area.active_tabwidget = self

        # callback to editor_area's public `create_empty_widget` Callable trait
        empty_widget = self.editor_area.create_empty_widget()

        self.addTab(empty_widget, '')
        self.empty_widget = empty_widget
        self.setFocus()

        # don't allow tab closing if empty widget comes up on a root tabwidget
        if self.parent().is_root():
            self.setTabsClosable(False)

        self.setTabText(0, '     ')

    def hide_empty_widget(self):
        """ Hides the empty widget (containing buttons to open new file, and
        collapse the split) based on whether the tabwidget is empty or not.
        """
        index = self.indexOf(self.empty_widget)
        self.removeTab(index)
        self.empty_widget.deleteLater()
        self.empty_widget = None
        self.setTabsClosable(True)

    def create_empty_widget(self):
        """ Creates the QFrame object to be shown when the current tabwidget is
        empty.
        """
        frame = QtGui.QFrame(parent=self)
        frame.setFrameShape(QtGui.QFrame.StyledPanel)
        layout = QtGui.QVBoxLayout(frame)

        # Add new file button and open file button only if the `callbacks` trait
        # of the editor_area has a callable for key `new` and key `open`
        new_file_action = self.editor_area.callbacks.get('new', None)
        open_file_action = self.editor_area.callbacks.get('open_dialog', None)
        open_show_dialog = False
        if open_file_action is None:
            open_file_action = self.editor_area.callbacks.get('open', None)
            open_show_dialog = True
        if not (new_file_action and open_file_action):
            return frame

        layout.addStretch()

        # generate new file button
        newfile_btn = QtGui.QPushButton('Create a new file', parent=frame)
        newfile_btn.clicked.connect(new_file_action)
        layout.addWidget(newfile_btn, alignment=QtCore.Qt.AlignHCenter)

        # generate label
        label = QtGui.QLabel(parent=frame)
        label.setText("""<span style='font-size:14px; color:#999999'>
                        or
                        </span>""")
        layout.addWidget(label, alignment=QtCore.Qt.AlignHCenter)

        # generate open button
        open_btn = QtGui.QPushButton('Select files from your computer', parent=frame)
        def _open():
            if open_show_dialog:
                open_dlg = FileDialog(action='open')
                open_dlg.open()
                self.editor_area.active_tabwidget = self
                if open_dlg.return_code == OK:
                    open_file_action(open_dlg.path)
            else:
                open_file_action()
        open_btn.clicked.connect(_open)
        layout.addWidget(open_btn, alignment=QtCore.Qt.AlignHCenter)

        # generate label
        label = QtGui.QLabel(parent=frame)
        label.setText("""<span style='font-size:14px; color:#999999'>
                        Tip: You can also drag and drop files/tabs here.
                        </span>""")
        layout.addWidget(label, alignment=QtCore.Qt.AlignHCenter)

        layout.addStretch()
        frame.setLayout(layout)

        return frame

    def get_names(self):
        """ Utility function to return names of all the editors open in the
        current tabwidget.
        """
        names = []
        for i in range(self.count()):
            editor_widget = self.widget(i)
            editor = self.editor_area._get_editor(editor_widget)
            if editor:
                names.append(editor.name)
        return names

    ###### Signal handlers ####################################################

    def _close_requested(self, index):
        """ Re-implemented to close the editor when it's tab is closed
        """
        widget = self.widget(index)

        # if close requested on empty_widget, collapse the pane and return
        if widget is self.empty_widget:
            self.parent().collapse()
            return
        editor = self.editor_area._get_editor(widget)
        editor.close()

    def _current_changed(self, index):
        """Re-implemented to update active editor
        """
        self.setCurrentIndex(index)
        editor_widget = self.widget(index)
        self.editor_area.active_editor = self.editor_area._get_editor(editor_widget)

    def tabInserted(self, index):
        """ Re-implemented to hide empty_widget when adding a new widget
        """
        # sets tab tooltip only if a real editor was added (not an empty_widget)
        editor = self.editor_area._get_editor(self.widget(index))
        if editor:
            self.setTabToolTip(index, editor.tooltip)

        if self.empty_widget:
            self.hide_empty_widget()

    def tabRemoved(self, index):
        """ Re-implemented to show empty_widget again if all tabs are removed
        """
        if not self.count() and not self.empty_widget:
            self.show_empty_widget()

    ##### Event handlers ######################################################

    def contextMenuEvent(self, event):
        """ To show collapse context menu even on empty tabwidgets
        """
        global_pos = self.mapToGlobal(event.pos())
        menu = self.editor_area.get_context_menu(pos=global_pos)
        qmenu = menu.create_menu(self)
        qmenu.exec_(global_pos)

    def dragEnterEvent(self, event):
        """ Re-implemented to highlight the tabwidget on drag enter
        """
        for handler in self.editor_area._all_drop_handlers:
            if handler.can_handle_drop(event, self):
                self.editor_area.active_tabwidget = self
                self.setBackgroundRole(QtGui.QPalette.Highlight)
                event.acceptProposedAction()
                return

        super(DraggableTabWidget, self).dragEnterEvent(event)

    def dropEvent(self, event):
        """ Re-implemented to handle drop events
        """
        for handler in self.editor_area._all_drop_handlers:
            if handler.can_handle_drop(event, self):
                handler.handle_drop(event, self)
                self.setBackgroundRole(QtGui.QPalette.Window)
                event.acceptProposedAction()
                break

    def dragLeaveEvent(self, event):
        """ Clear widget highlight on leaving
        """
        self.setBackgroundRole(QtGui.QPalette.Window)
        return super(DraggableTabWidget, self).dragLeaveEvent(event)


class DraggableTabBar(QtGui.QTabBar):
    """ Implements a QTabBar with event filters for tab drag
    """
    def __init__(self, editor_area, parent):
        super(DraggableTabBar, self).__init__(parent)
        self.editor_area = editor_area
        self.setContextMenuPolicy(QtCore.Qt.DefaultContextMenu)
        self.drag_obj = None

    def mousePressEvent(self, event):
        if event.button()==QtCore.Qt.LeftButton:
            index = self.tabAt(event.pos())
            tabwidget = self.parent()
            if tabwidget.widget(index) and (not tabwidget.widget(index) == tabwidget.empty_widget):
                self.drag_obj = TabDragObject(start_pos=event.pos(), tabBar=self)
        return super(DraggableTabBar, self).mousePressEvent(event)

    def mouseMoveEvent(self, event):
        """ Re-implemented to create a drag event when the mouse is moved for a
        sufficient distance while holding down mouse button.
        """
        # go into the drag logic only if a drag_obj is active
        if self.drag_obj:
            # is the left mouse button still pressed?
            if not event.buttons()==QtCore.Qt.LeftButton:
                pass
            # has the mouse been dragged for sufficient distance?
            elif ((event.pos() - self.drag_obj.start_pos).manhattanLength()
                < QtGui.QApplication.startDragDistance()):
                pass
            # initiate drag
            else:
                drag = QtGui.QDrag(self.drag_obj.widget)
                mimedata = PyMimeData(data=self.drag_obj, pickle=False)
                drag.setPixmap(self.drag_obj.get_pixmap())
                drag.setHotSpot(self.drag_obj.get_hotspot())
                drag.setMimeData(mimedata)
                drag.exec_()
                self.drag_obj = None # deactivate the drag_obj again
                return
        return super(DraggableTabBar, self).mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        """ Re-implemented to deactivate the drag when mouse button is
        released
        """
        self.drag_obj = None
        return super(DraggableTabBar, self).mouseReleaseEvent(event)


class TabDragObject(object):
    """ Class to hold information related to tab dragging/dropping
    """

    def __init__(self, start_pos, tabBar):
        """
        Parameters
        ----------

        start_pos : position in tabBar coordinates where the drag was started
        tabBar : tabBar containing the tab on which drag started
        """
        self.start_pos = start_pos
        self.from_index = tabBar.tabAt(self.start_pos)
        self.from_editor_area = tabBar.parent().editor_area
        self.widget = tabBar.parent().widget(self.from_index)
        self.from_tabbar = tabBar

    def get_pixmap(self):
        """ Returns the drag pixmap including page widget and tab rectangle.
        """
        # instatiate the painter object with gray-color filled pixmap
        tabBar = self.from_tabbar
        tab_rect = tabBar.tabRect(self.from_index)
        size = self.widget.rect().size()
        result_pixmap = QtGui.QPixmap(size)
        painter = QtGui.QStylePainter(result_pixmap, tabBar)

        painter.fillRect(result_pixmap.rect(), QtCore.Qt.lightGray)
        painter.setCompositionMode(QtGui.QPainter.CompositionMode_SourceOver)

        optTabBase = QtGui.QStyleOptionTabBarBaseV2()
        optTabBase.initFrom(tabBar)
        painter.drawPrimitive(QtGui.QStyle.PE_FrameTabBarBase, optTabBase)

        # region of active tab
        pixmap1 = QtGui.QPixmap.grabWidget(tabBar, tab_rect)
        painter.drawPixmap(0,0,pixmap1) #tab_rect.topLeft(), pixmap1)

        # region of the page widget
        pixmap2 = QtGui.QPixmap.grabWidget(self.widget)
        painter.drawPixmap(0, tab_rect.height(), size.width(), size.height(), pixmap2)

        # finish painting
        painter.end()

        return result_pixmap

    def get_hotspot(self):
        return self.start_pos - self.from_tabbar.tabRect(self.from_index).topLeft()

###############################################################################
# Default drop handlers.
###############################################################################

class TabDropHandler(BaseDropHandler):
    """ Class to handle tab drop events
    """

    # whether to allow dragging of tabs across different opened windows
    allow_cross_window_drop = Bool(False)

    def can_handle_drop(self, event, target):
        if isinstance(event.mimeData(), PyMimeData) and \
            isinstance(event.mimeData().instance(), TabDragObject):
            if not self.allow_cross_window_drop:
                drag_obj = event.mimeData().instance()
                return drag_obj.from_editor_area == target.editor_area
            else:
                return True
        return False

    def handle_drop(self, event, target):
        # get the drop object back
        drag_obj = event.mimeData().instance()

        # extract widget label
        # (editor_area is common to both source and target in most cases but when
        # the dragging happens across different windows, they are not, and hence it
        # must be pulled in directly from the source)
        editor = target.editor_area._get_editor(drag_obj.widget)
        label = target.editor_area._get_label(editor)

        # if drop occurs at a tab bar, insert the tab at that position
        if not target.tabBar().tabAt(event.pos())==-1:
            index = target.tabBar().tabAt(event.pos())
            target.insertTab(index, drag_obj.widget, label)

        else:
            # if the drag initiated from the same tabwidget, put the tab
            # back at the original index
            if target is drag_obj.from_tabbar.parent():
                target.insertTab(drag_obj.from_index, drag_obj.widget, label)
            # else, just add it at the end
            else:
                target.addTab(drag_obj.widget, label)

        # make the dropped widget active
        target.setCurrentWidget(drag_obj.widget)



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms described in the PyQt GPL exception also apply

#
# Author: Riverbank Computing Limited
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" The base class for all pyface wizards. """


# Major package imports.
from pyface.qt import QtCore, QtGui

# Enthought library imports.
from traits.api import Bool, Instance, List, Property, provides, Unicode
from pyface.api import Dialog
from pyface.wizard.i_wizard import IWizard, MWizard
from pyface.wizard.i_wizard_controller import IWizardController
from pyface.wizard.i_wizard_page import IWizardPage


@provides(IWizard)
class Wizard(MWizard, Dialog):
    """ The base class for all pyface wizards.

    See the IWizard interface for the API documentation.

    """


    #### 'IWizard' interface ##################################################

    pages = Property(List(IWizardPage))

    controller = Instance(IWizardController)

    show_cancel = Bool(True)

    #### 'IWindow' interface ##################################################

    title = Unicode('Wizard')

    ###########################################################################
    # 'IWizard' interface.
    ###########################################################################

    # Override MWizard implementation to do nothing. We still call these methods
    # because it expected by IWizard, and users may wish to hook in custom code
    # before changing a page.

    def next(self):
        pass

    def previous(self):
        pass

    ###########################################################################
    # Protected 'IDialog' interface.
    ###########################################################################

    def _create_contents(self, parent):
        pass

    ###########################################################################
    # Protected 'IWidget' interface.
    ###########################################################################

    def _create_control(self, parent):
        control = _Wizard(parent, self)
        control.setOptions(QtGui.QWizard.NoDefaultButton |
                           QtGui.QWizard.NoBackButtonOnStartPage)
        control.setModal(self.style == 'modal')
        control.setWindowTitle(self.title)

        # Necessary for 'nonmodal'. See Dialog for more info.
        if self.style == 'nonmodal':
            control.finished.connect(self._finished_fired)

        if self.size != (-1, -1):
            size = QtCore.QSize(*self.size)
            control.setMinimumSize(size)
            control.resize(size)

        if not self.show_cancel:
            control.setOption(QtGui.QWizard.NoCancelButton)

        if self.help_id:
            control.setOption(QtGui.QWizard.HaveHelpButton)
            control.helpRequested.connect(self._help_requested)

        # Add the initial pages.
        for page in self.pages:
            page.pyface_wizard = self
            control.addWizardPage(page)

        # Set the start page.
        control.setStartWizardPage()

        return control

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _help_requested(self):
        """ Called when the 'Help' button is pressed. """

        # FIXME: Hook into a help system.
        print("Show help for", self.help_id)

    #### Trait handlers #######################################################

    def _get_pages(self):
        """ The pages getter. """

        return self.controller.pages

    def _set_pages(self, pages):
        """ The pages setter. """

        # Remove pages from the old list that appear in the new list.  The old
        # list will now contain pages that are no longer in the wizard.
        old_pages = self.pages
        new_pages = []

        for page in pages:
            try:
                old_pages.remove(page)
            except ValueError:
                new_pages.append(page)

        # Dispose of the old pages.
        for page in old_pages:
            page.dispose_page()

        # If we have created the control then we need to add the new pages,
        # otherwise we leave it until the control is created.
        if self.control:
            for page in new_pages:
                self.control.addWizardPage(page)

        self.controller.pages = pages

    def _controller_default(self):
        """ Provide a default controller. """

        from pyface.wizard.wizard_controller import WizardController

        return WizardController()


class _Wizard(QtGui.QWizard):
    """ A QWizard sub-class that hooks into the controller to determine the
    next page to show. """

    def __init__(self, parent, pyface_wizard):
        """ Initialise the object. """

        QtGui.QWizard.__init__(self, parent)

        self._pyface_wizard = pyface_wizard
        self._controller = pyface_wizard.controller
        self._ids = {}

        self.currentIdChanged.connect(self._update_controller)

    def addWizardPage(self, page):
        """ Add a page that provides IWizardPage. """

        # We must pass a parent otherwise TraitsUI does the wrong thing.
        qpage = page.create_page(self)
        qpage.pyface_wizard = self._pyface_wizard
        id = self.addPage(qpage)
        self._ids[id] = page

    def setStartWizardPage(self):
        """ Set the first page. """

        page = self._controller.get_first_page()
        id = self._page_to_id(page)

        if id >= 0:
            self.setStartId(id)

    def nextId(self):
        """ Reimplemented to return the id of the next page to display. """

        if self.currentId() < 0:
            return self._page_to_id(self._controller.get_first_page())
        current = self._ids[self.currentId()]
        next = self._controller.get_next_page(current)

        return self._page_to_id(next)

    def _update_controller(self, id):
        """ Called when the current page has changed. """

        # Keep the controller in sync with the wizard.
        self._controller.current_page = self._ids.get(id)

    def _page_to_id(self, page):
        """ Return the id of the given page. """

        if page is None:
            id = -1
        else:
            for id, p in list(self._ids.items()):
                if p is page:
                    break
            else:
                id = -1

        return id

#### EOF ######################################################################



""" A simple list box widget with a model-view architecture. """


# Major package imports.
import wx

# Enthought library imports.
from traits.api import Event, Instance, Int

# Local imports.
from pyface.list_box_model import ListBoxModel
from .widget import Widget


class ListBox(Widget):
    """ A simple list box widget with a model-view architecture. """

    # The model that provides the data for the list box.
    model = Instance(ListBoxModel)

    # The objects currently selected in the list.
    selection = Int(-1)

    # Events.

    # An item has been activated.
    item_activated = Event

    # Default style.
    STYLE = wx.LB_SINGLE | wx.LB_HSCROLL | wx.LB_NEEDED_SB


    def __init__(self, parent, **traits):
        """ Creates a new list box. """

        # Base-class constructors.
        super(ListBox, self).__init__(**traits)

        # Create the widget!
        self._create_control(parent)

        # Listen for changes to the model.
        self.model.on_trait_change(self._on_model_changed, "list_changed")

        return

    def dispose(self):
        self.model.on_trait_change(self._on_model_changed, "list_changed",
                                   remove = True)
        self.model.dispose()
        return

    ###########################################################################
    # 'ListBox' interface.
    ###########################################################################

    def refresh(self):
        """ Refreshes the list box. """

        # For now we just clear out the entire list.
        self.control.Clear()

        # Populate the list.
        self._populate()

        return

    ###########################################################################
    # wx event handlers.
    ###########################################################################

    def _on_item_selected(self, event):
        """ Called when an item in the list is selected. """

        listbox = event.GetEventObject()

        self.selection = listbox.GetSelection()

        return

    def _on_item_activated(self, event):
        """ Called when an item in the list is activated. """

        listbox = event.GetEventObject()
        index = listbox.GetSelection()

        # Trait event notification.
        self.item_activated = index

        return

    ###########################################################################
    # Trait handlers.
    ###########################################################################

    #### Static ###############################################################

    def _selection_changed(self, index):
        """ Called when the selected item is changed. """

        if index != -1:
            self.control.SetSelection(index)

        return

    #### Dynamic ##############################################################

    def _on_model_changed(self, event):
        """ Called when the model has changed. """

        # For now we just clear out the entire list.
        self.refresh()

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _create_control(self, parent):
        """ Creates the widget. """

        self.control = wx.ListBox(parent, -1, style = self.STYLE)

        # Wire it up!
        wx.EVT_LISTBOX(self.control, self.control.GetId(),
                       self._on_item_selected)
        wx.EVT_LISTBOX_DCLICK(self.control, self.control.GetId(),
                              self._on_item_activated)

        # Populate the list.
        self._populate()

        return

    def _populate(self):
        """ Populates the list box. """

        for index in range(self.model.get_item_count()):
            label, item = self.model.get_item_at(index)
            self.control.Append(label, item)

        return

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from traits.api import Any, HasTraits, provides

# Local imports.
from pyface.i_widget import IWidget, MWidget


@provides(IWidget)
class Widget(MWidget, HasTraits):
    """ The toolkit specific implementation of a Widget.  See the IWidget
    interface for the API documentation.
    """


    #### 'IWidget' interface ##################################################

    control = Any

    parent = Any

    ###########################################################################
    # 'IWidget' interface.
    ###########################################################################

    def destroy(self):
        self.control = None

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" The abstract interface for all pyface dialogs. """


# Enthought library imports.
from traits.api import Bool, Enum, Int, Str, Unicode

# Local imports.
from pyface.constant import OK
from pyface.i_window import IWindow


class IDialog(IWindow):
    """ The abstract interface for all pyface dialogs.

    Usage: Sub-class this class and either override '_create_contents' or
    more simply, just override the two methods that do the real work:-

    1) '_create_dialog_area' creates the main content of the dialog.
    2) '_create_buttons'     creates the dialog buttons.
    """

    #### 'IDialog' interface ##################################################

    #: The label for the 'cancel' button.  The default is toolkit specific.
    cancel_label = Unicode

    #: The context sensitive help Id (the 'Help' button is only shown iff this
    #: is set).
    help_id = Str

    #: The label for the 'help' button.  The default is toolkit specific.
    help_label = Unicode

    #: The label for the 'ok' button.  The default is toolkit specific.
    ok_label = Unicode

    #: Is the dialog resizeable?
    resizeable = Bool(True)

    #: The return code after the window is closed to indicate whether the dialog
    #: was closed via 'Ok' or 'Cancel').
    return_code = Int(OK)

    #: The dialog style (is it modal or not).
    # FIXME v3: It doesn't seem possible to use non-modal dialogs.  (How do you
    # get access to the buttons?)
    style = Enum('modal', 'nonmodal')

    ###########################################################################
    # 'IDialog' interface.
    ###########################################################################

    def open(self):
        """ Opens the dialog.

        If the dialog is modal then the dialog's event loop is entered and the
        dialog closed afterwards.  The 'return_code' trait is updated according
        to the button the user pressed and this value is also returned.

        If the dialog is non-modal the return_code trait is set to 'OK'.

        Returns
        -------
        return_code : OK or CANCEL
            The value of the ``return_code`` trait.
        """

    ###########################################################################
    # Protected 'IDialog' interface.
    ###########################################################################

    def _create_buttons(self, parent):
        """ Create and return the buttons.

        Parameters
        ----------
        parent : toolkit control
            The dialog's toolkit control to be used as the parent for
            buttons.

        Returns
        -------
        buttons : toolkit control
            A control containing the dialog's buttons.

        """

    def _create_contents(self, parent):
        """ Create and return the dialog's contents.

        Parameters
        ----------
        parent : toolkit control
            The window's toolkit control to be used as the parent for
            widgets in the contents.

        Returns
        -------
        control : toolkit control
            A control to be used for contents of the window.
        """

    def _create_dialog_area(self, parent):
        """ Create and return the main content of the dialog's window.

        Parameters
        ----------
        parent : toolkit control
            A toolkit control to be used as the parent for widgets in the
            contents.

        Returns
        -------
        control : toolkit control
            A control to be used for main contents of the dialog.
        """

    def _show_modal(self):
        """ Opens the dialog as a modal dialog.

        Returns
        -------
        return_code : OK or CANCEL
            The return code from the user's interactions.
        """


class MDialog(object):
    """ The mixin class that contains common code for toolkit specific
    implementations of the IDialog interface.

    Implements: open()
    Reimplements: _add_event_listeners(), _create()
    """

    ###########################################################################
    # 'IDialog' interface.
    ###########################################################################

    def open(self):
        """ Opens the dialog.

        If the dialog is modal then the dialog's event loop is entered and the
        dialog closed afterwards.  The 'return_code' trait is updated according
        to the button the user pressed and this value is also returned.

        If the dialog is non-modal the return_code trait is set to 'OK'.

        Returns
        -------
        return_code : OK or CANCEL
            The value of the ``return_code`` trait.
        """
        if self.control is None:
            self._create()

        if self.style == 'modal':
            self.return_code = self._show_modal()
            self.close()

        else:
            self.show(True)
            self.return_code = OK

        return self.return_code

    ###########################################################################
    # Protected 'IWidget' interface.
    ###########################################################################

    def _create(self):
        """ Creates the window's widget hierarchy. """

        super(MDialog, self)._create()

        self._create_contents(self.control)

    ###########################################################################
    # Protected 'IWindow' interface.
    ###########################################################################

    def _add_event_listeners(self):
        """ Adds any event listeners required by the window. """

        # We don't bother for dialogs.
        pass





# Local imports.
from .dock_pane_toggle_group import DockPaneToggleGroup
from .schema import ActionSchema, GroupSchema, MenuSchema, MenuBarSchema, \
    ToolBarSchema, SGroup, SMenu, SMenuBar, SToolBar
from .schema_addition import SchemaAddition
from .task_action import CentralPaneAction, DockPaneAction, EditorAction, \
     TaskAction, TaskWindowAction
from .task_action_controller import TaskActionController
from .task_action_manager_builder import TaskActionManagerBuilder
from .task_toggle_group import TaskToggleGroup



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" A tool bar manager realizes itself in a tool palette control. """


# Import the toolkit specific version.
from pyface.toolkit import toolkit_object
ToolPaletteManager = toolkit_object('action.tool_palette_manager:ToolPaletteManager')

### EOF #######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" The base interface for all pyface widgets. """


# Enthought library imports.
from traits.api import Any, Bool, HasTraits, Interface


class IWidget(Interface):
    """ The base interface for all pyface widgets.

    Pyface widgets delegate to a toolkit specific control.
    """

    #: The toolkit specific control that represents the widget.
    control = Any

    #: The control's optional parent control.
    parent = Any

    #: Whether or not the control is visible
    visible = Bool(True)

    #: Whether or not the control is enabled
    enabled = Bool(True)

    # ------------------------------------------------------------------------
    # 'IWidget' interface.
    # ------------------------------------------------------------------------

    def show(self, visible):
        """ Show or hide the widget.

        Parameter
        ---------
        visible : bool
            Visible should be ``True`` if the widget should be shown.
        """

    def enable(self, enabled):
        """ Enable or disable the widget.

        Parameter
        ---------
        enabled : bool
            The enabled state to set the widget to.
        """

    def destroy(self):
        """ Destroy the control if it exists. """

    # ------------------------------------------------------------------------
    # Protected 'IWidget' interface.
    # ------------------------------------------------------------------------

    def _create(self):
        """ Creates the toolkit specific control.

        This method should create the control and assign it to the
        :py:attr:``control`` trait.
        """

    def _create_control(self, parent):
        """ Create toolkit specific control that represents the widget.

        Parameters
        ----------
        parent : toolkit control
            The toolkit control to be used as the parent for the widget's
            control.

        Returns
        -------
        control : toolkit control
            A control for the widget.
        """

    def _add_event_listeners(self):
        """ Set up toolkit-specific bindings for events """

    def _remove_event_listeners(self):
        """ Remove toolkit-specific bindings for events """


class MWidget(object):
    """ The mixin class that contains common code for toolkit specific
    implementations of the IWidget interface.
    """

    # ------------------------------------------------------------------------
    # Protected 'IWidget' interface.
    # ------------------------------------------------------------------------

    def _create(self):
        """ Creates the toolkit specific control.

        This method should create the control and assign it to the
        :py:attr:``control`` trait.
        """
        self.control = self._create_control(self.parent)
        self._add_event_listeners()

    def _create_control(self, parent):
        """ Create toolkit specific control that represents the widget.

        Parameters
        ----------
        parent : toolkit control
            The toolkit control to be used as the parent for the widget's
            control.

        Returns
        -------
        control : toolkit control
            A control for the widget.
        """
        raise NotImplementedError

    def _add_event_listeners(self):
        """ Set up toolkit-specific bindings for events """
        pass

    def _remove_event_listeners(self):
        """ Remove toolkit-specific bindings for events """
        pass



""" The node type for a trait dictionary. """


# Enthought library imports.
from traits.api import Any, Str

# Local imports.
from .node_type import NodeType


class TraitDictNodeType(NodeType):
    """ The node type for a trait dictionary. """

    #### 'TraitDictNodeType' interface ########################################

    # The type of object that provides the trait dictionary.
    klass = Any

    # The label text.
    text = Str

    # The trait name.
    trait_name = Str

    ###########################################################################
    # 'NodeType' interface.
    ###########################################################################

    def is_type_for(self, node):
        """ Returns True if this node type recognizes a node. """

        is_type_for = isinstance(node, dict) \
                      and hasattr(node, 'object') \
                      and isinstance(node.object, self.klass) \
                      and node.name == self.trait_name

        return is_type_for

    def allows_children(self, node):
        """ Does the node allow children (ie. a folder vs a file). """

        return True

    def has_children(self, node):
        """ Returns True if a node has children, otherwise False. """

        return len(node) > 0

    def get_children(self, node):
        """ Returns the children of a node. """

        return list(node.values())

    def get_text(self, node):
        """ Returns the label text for a node. """

        return self.text

##### EOF #####################################################################



""" The editor manager interface. """


# Enthought library imports.
from traits.api import Instance, Interface


class IEditorManager(Interface):
    """ The editor manager interface. """

    # The workbench window that the editor manager manages editors for ;^)
    window = Instance('pyface.workbench.api.WorkbenchWindow')

    def add_editor(self, editor, kind):
        """ Registers an existing editor. """

    def create_editor(self, window, obj, kind):
        """ Create an editor for an object.

        'kind' optionally contains any data required by the specific editor
        manager implementation to decide what type of editor to create.

        Returns None if no editor can be created for the resource.
        """

    def get_editor(self, window, obj, kind):
        """ Get the editor that is currently editing an object.

        'kind' optionally contains any data required by the specific editor
        manager implementation to decide what type of editor to create.

        Returns None if no such editor exists.
        """

    def get_editor_kind(self, editor):
        """ Return the 'kind' associated with 'editor'. """

    def get_editor_memento(self, editor):
        """ Return the state of an editor suitable for pickling etc.
        """

    def set_editor_memento(self, memento):
        """ Restore an editor from a memento and return it. """

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" The interface for all pyface wizard controllers. """


# Enthought library imports.
from traits.api import Bool, Interface, Instance, List

# Local imports.
from .i_wizard_page import IWizardPage


class IWizardController(Interface):
    """ The interface for all pyface wizard controllers. """

    #### 'IWizardController' interface ########################################

    # The pages under the control of this controller.
    pages = List(IWizardPage)

    # The current page.
    current_page = Instance(IWizardPage)

    # Set if the wizard complete.
    complete = Bool(False)

    ###########################################################################
    # 'IWizardController' interface.
    ###########################################################################

    def get_first_page(self):
        """ Returns the first page in the model. """

    def get_next_page(self, page):
        """ Returns the next page. """

    def get_previous_page(self, page):
        """ Returns the previous page. """

    def is_first_page(self, page):
        """ Is the page the first page? """

    def is_last_page(self, page):
        """ Is the page the last page? """

    def dispose_pages(self):
        """ Dispose all the pages. """

#### EOF ######################################################################



""" The default action controller for menus, menu bars and tool bars. """


# Enthought library imports.
from traits.api import HasTraits


class ActionController(HasTraits):
    """ The default action controller for menus, menu bars and tool bars. """

    ###########################################################################
    # 'ActionController' interface.
    ###########################################################################

    def perform(self, action, event):
        """ Control an action invocation.

        Parameters
        ----------
        action : Action instance
            The action to perform.
        event : ActionEvent instance
            The event that triggered the action.

        Returns
        -------
        result : any
            The result of the action's perform method (usually None).
        """
        return action.perform(event)

    def can_add_to_menu(self, action):
        """ Can add an action to a menu

        Parameters
        ----------
        action : Action instance
            The action to consider.

        Returns
        -------
        can_add : bool
            ``True` if the action can be added to a menu/menubar.
        """
        return True

    def add_to_menu(self, action):
        """ Called when an action is added to the a menu/menubar.

        Parameters
        ----------
        action : Action instance
            The action added to the menu.
        """
        pass

    def can_add_to_toolbar(self, action):
        """ Returns True if the action can be added to a toolbar.

        Parameters
        ----------
        action : Action instance
            The action to consider.

        Returns
        -------
        can_add : bool
            ``True` if the action can be added to a toolbar.
        """
        return True

    def add_to_toolbar(self, action):
        """ Called when an action is added to the a toolbar.

        Parameters
        ----------
        action : Action instance
            The action added to the toolbar.
        """
        pass



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------



from .action import Action
from .action_controller import ActionController
from .action_event import ActionEvent
from .action_item import ActionItem
from .action_manager import ActionManager
from .action_manager_item import ActionManagerItem
from .group import Group, Separator
from .menu_manager import MenuManager
from .menu_bar_manager import MenuBarManager
from .status_bar_manager import StatusBarManager
from .tool_bar_manager import ToolBarManager
from .window_action import WindowAction


###############################################################################
# This part of the module handles widgets that are still wx specific.  This
# will all be removed when everything has been ported to PyQt and pyface
# becomes toolkit agnostic.
###############################################################################

from traits.etsconfig.api import ETSConfig
if ETSConfig.toolkit == 'wx':
    from .tool_palette_manager import ToolPaletteManager





import platform

from traits.testing.unittest_tools import unittest

from ..dialog import Dialog
from ..constant import OK, CANCEL
from ..toolkit import toolkit_object


is_qt = toolkit_object.toolkit == 'qt4'
if is_qt:
    from pyface.qt import qt_api

GuiTestAssistant = toolkit_object('util.gui_test_assistant:GuiTestAssistant')
no_gui_test_assistant = (GuiTestAssistant.__name__ == 'Unimplemented')

ModalDialogTester = toolkit_object('util.modal_dialog_tester:ModalDialogTester')
no_modal_dialog_tester = (ModalDialogTester.__name__ == 'Unimplemented')

is_pyqt5 = (is_qt and qt_api == 'pyqt5')
is_pyqt4_linux = (is_qt and qt_api == 'pyqt' and platform.system() == 'Linux')


@unittest.skipIf(no_gui_test_assistant, 'No GuiTestAssistant')
class TestDialog(unittest.TestCase, GuiTestAssistant):

    def setUp(self):
        GuiTestAssistant.setUp(self)
        self.dialog = Dialog()

    def tearDown(self):
        if self.dialog.control is not None:
            with self.delete_widget(self.dialog.control):
                self.dialog.destroy()
        self.dialog = None
        GuiTestAssistant.tearDown(self)

    def test_create(self):
        # test that creation and destruction works as expected
        self.dialog._create()
        self.event_loop()
        self.dialog.destroy()
        self.event_loop()

    def test_destroy(self):
        # test that destroy works even when no control
        self.dialog.destroy()
        self.event_loop()

    def test_size(self):
        # test that size works as expected
        self.dialog.size = (100, 100)
        self.dialog._create()
        self.event_loop()
        self.dialog.destroy()
        self.event_loop()

    def test_position(self):
        # test that position works as expected
        self.dialog.position = (100, 100)
        self.dialog._create()
        self.event_loop()
        self.dialog.destroy()
        self.event_loop()

    def test_create_ok_renamed(self):
        # test that creation and destruction works as expected with ok_label
        self.dialog.ok_label = "Sure"
        self.dialog._create()
        self.event_loop()
        self.dialog.destroy()
        self.event_loop()

    def test_create_cancel_renamed(self):
        # test that creation and destruction works as expected with cancel_label
        self.dialog.cancel_label = "I Don't Think So"
        self.dialog._create()
        self.event_loop()
        self.dialog.destroy()
        self.event_loop()

    def test_create_help(self):
        # test that creation and destruction works as expected with help
        self.dialog.help_id = "test_help"
        self.dialog._create()
        self.event_loop()
        self.dialog.destroy()

    def test_create_help_label(self):
        # test that creation and destruction works as expected with help
        self.dialog.help_id = "test_help"
        self.dialog.help_label = "Assistance"
        self.dialog._create()
        self.event_loop()
        self.dialog.destroy()
        self.event_loop()

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_accept(self):
        # test that accept works as expected
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_run(when_opened=lambda x: x.close(accept=True))
        self.assertEqual(tester.result, OK)
        self.assertEqual(self.dialog.return_code, OK)

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_reject(self):
        # test that reject works as expected
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_run(when_opened=lambda x: x.close(accept=False))
        self.assertEqual(tester.result, CANCEL)
        self.assertEqual(self.dialog.return_code, CANCEL)

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_close(self):
        # test that closing works as expected
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_run(when_opened=lambda x: self.dialog.close())
        self.assertEqual(tester.result, CANCEL)
        self.assertEqual(self.dialog.return_code, CANCEL)

    @unittest.skipIf(is_pyqt5, "Dialog click tests don't work on pyqt5.")  # noqa
    @unittest.skipIf(is_pyqt4_linux, "Dialog click tests don't work reliably on linux.  Issue #282.")  # noqa
    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_ok(self):
        # test that OK works as expected
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_wait(when_opened=lambda x: x.click_button(OK))
        self.assertEqual(tester.result, OK)
        self.assertEqual(self.dialog.return_code, OK)

    @unittest.skipIf(is_pyqt5, "Dialog click tests don't work on pyqt5.")  # noqa
    @unittest.skipIf(is_pyqt4_linux, "Dialog click tests don't work reliably on linux.  Issue #282.")  # noqa
    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_cancel(self):
        # test that cancel works as expected
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_run(when_opened=lambda x: x.click_button(CANCEL))
        self.assertEqual(tester.result, CANCEL)
        self.assertEqual(self.dialog.return_code, CANCEL)

    @unittest.skipIf(is_pyqt5, "Dialog click tests don't work on pyqt5.")  # noqa
    @unittest.skipIf(is_pyqt4_linux, "Dialog click tests don't work reliably on linux.  Issue #282.")  # noqa
    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_renamed_ok(self):
        self.dialog.ok_label = "Sure"
        # test that OK works as expected if renames
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_wait(when_opened=lambda x: x.click_widget("Sure"))
        self.assertEqual(tester.result, OK)
        self.assertEqual(self.dialog.return_code, OK)

    @unittest.skipIf(is_pyqt5, "Dialog click tests don't work on pyqt5.")  # noqa
    @unittest.skipIf(is_pyqt4_linux, "Dialog click tests don't work reliably on linux.  Issue #282.")  # noqa
    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_renamed_cancel(self):
        self.dialog.cancel_label = "I Don't Think So"
        # test that OK works as expected if renames
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_wait(when_opened=lambda x: x.click_widget("I Don't Think So"))
        self.assertEqual(tester.result, CANCEL)
        self.assertEqual(self.dialog.return_code, CANCEL)

    @unittest.skipIf(is_pyqt5, "Dialog click tests don't work on pyqt5.")  # noqa
    @unittest.skipIf(is_pyqt4_linux, "Dialog click tests don't work reliably on linux.  Issue #282.")  # noqa
    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_help(self):
        def click_help_and_close(tester):
            tester.click_widget("Help")
            tester.close(accept=True)

        self.dialog.help_id = "help_test"
        # test that OK works as expected if renames
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_wait(when_opened=click_help_and_close)
        self.assertEqual(tester.result, OK)
        self.assertEqual(self.dialog.return_code, OK)

    @unittest.skipIf(is_pyqt5, "Dialog click tests don't work on pyqt5.")  # noqa
    @unittest.skipIf(is_pyqt4_linux, "Dialog click tests don't work reliably on linux.  Issue #282.")  # noqa
    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_renamed_help(self):
        def click_help_and_close(tester):
            tester.click_widget("Assistance")
            tester.close(accept=True)

        self.dialog.help_id = "help_test"
        self.dialog.help_label = "Assistance"
        # test that OK works as expected if renames
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_wait(when_opened=click_help_and_close)
        self.assertEqual(tester.result, OK)
        self.assertEqual(self.dialog.return_code, OK)

    def test_nonmodal_close(self):
        # test that closing works as expected
        self.dialog.style = 'nonmodal'
        result = self.dialog.open()
        self.event_loop()
        self.dialog.close()
        self.event_loop()
        self.assertEqual(result, OK)
        self.assertEqual(self.dialog.return_code, OK)

    def test_not_resizable(self):
        # test that a resizable dialog can be created
        # XXX use nonmodal for better cross-platform coverage
        self.dialog.style = 'nonmodal'
        self.dialog.resizable = False
        result = self.dialog.open()
        self.event_loop()
        self.dialog.close()
        self.event_loop()
        self.assertEqual(result, OK)
        self.assertEqual(self.dialog.return_code, OK)





from traits.testing.unittest_tools import unittest

from ..progress_dialog import ProgressDialog
from ..util.gui_test_assistant import GuiTestAssistant


class TestProgressDialog(unittest.TestCase, GuiTestAssistant):

    def setUp(self):
        GuiTestAssistant.setUp(self)
        self.dialog = ProgressDialog()

    def tearDown(self):
        if self.dialog.control is not None:
            with self.delete_widget(self.dialog.control):
                self.dialog.destroy()
        GuiTestAssistant.tearDown(self)

    def test_create(self):
        # test that creation and destruction works as expected
        self.dialog._create()
        self.gui.process_events()
        self.assertIsNotNone(self.dialog.control)
        self.assertIsNotNone(self.dialog.progress_bar)
        self.assertIsNotNone(self.dialog._message_control)
        self.assertIsNone(self.dialog._elapsed_control)
        self.assertIsNone(self.dialog._estimated_control)
        self.assertIsNone(self.dialog._remaining_control)
        self.dialog.destroy()
        self.gui.process_events()

    def test_show_time(self):
        # test that creation works with show_time
        self.dialog.show_time =  True
        self.dialog._create()
        self.gui.process_events()
        self.assertIsNotNone(self.dialog._elapsed_control)
        self.assertIsNotNone(self.dialog._estimated_control)
        self.assertIsNotNone(self.dialog._remaining_control)
        self.dialog.destroy()
        self.gui.process_events()

    def test_show_percent(self):
        # test that creation works with show_percent
        self.dialog.show_percent =  True
        self.dialog._create()
        self.gui.process_events()
        self.assertEqual(self.dialog.progress_bar.format(), "%p%")
        self.dialog.destroy()
        self.gui.process_events()

    def test_update(self):
        self.dialog.min = 0
        self.dialog.max = 10
        self.dialog.open()
        for i in range(11):
            result = self.dialog.update(i)
            self.gui.process_events()
            self.assertEqual(result, (True, False))
            if i < 10:
                self.assertEqual(self.dialog.progress_bar.value(), i)
        self.assertIsNone(self.dialog.control)
        self.gui.process_events()

    def test_update_no_control(self):
        # note: inconsistent implementation with Wx
        self.dialog.min = 0
        self.dialog.max = 10
        result = self.dialog.update(1)
        self.assertEqual(result, (None, None))
        self.gui.process_events()

    def test_change_message(self):
        self.dialog.min = 0
        self.dialog.max = 10
        self.dialog.open()
        for i in range(11):
            self.dialog.change_message('Updating {}'.format(i))
            result = self.dialog.update(i)
            self.gui.process_events()
            self.assertEqual(result, (True, False))
            self.assertEqual(self.dialog.message, 'Updating {}'.format(i))
            self.assertEqual(self.dialog._message_control.text(),
                             'Updating {}'.format(i))
        self.assertIsNone(self.dialog.control)
        self.gui.process_events()

    def test_change_message_trait(self):
        self.dialog.min = 0
        self.dialog.max = 10
        self.dialog.open()
        for i in range(11):
            self.dialog.message = 'Updating {}'.format(i)
            result = self.dialog.update(i)
            self.gui.process_events()
            self.assertEqual(result, (True, False))
            self.assertEqual(self.dialog.message, 'Updating {}'.format(i))
            self.assertEqual(self.dialog._message_control.text(),
                             'Updating {}'.format(i))
        self.assertIsNone(self.dialog.control)
        self.gui.process_events()

    def test_update_show_time(self):
        self.dialog.min = 0
        self.dialog.max = 10
        self.dialog.show_time = True
        self.dialog.open()
        for i in range(11):
            result = self.dialog.update(i)
            self.gui.process_events()
            self.assertEqual(result, (True, False))
            self.assertNotEqual(self.dialog._elapsed_control.text(), "")
            self.assertNotEqual(self.dialog._estimated_control.text(), "")
            self.assertNotEqual(self.dialog._remaining_control.text(), "")
        self.assertIsNone(self.dialog.control)
        self.gui.process_events()



""" The event that is generated when a key is pressed. """


# Enthought library imports.
from traits.api import Bool, HasTraits, Int, Any


class KeyPressedEvent(HasTraits):
    """ The event that is generated when a key is pressed. """

    #### 'KeyPressedEvent' interface ##########################################

    #: Is the alt key down?
    alt_down = Bool

    #: Is the control key down?
    control_down = Bool

    #: Is the shift key down?
    shift_down = Bool

    #: The keycode.
    key_code = Int

    #: The original toolkit specific event.
    event = Any

#### EOF ######################################################################



#-------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   12/14/2005
#
#-------------------------------------------------------------------------------

""" Defines the IDockable interface which objects contained in a DockWindow
    DockControl can implement in order to allow themselves to be dragged into
    a different DockWindow.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import wx

#-------------------------------------------------------------------------------
#  'IDockable' class:
#-------------------------------------------------------------------------------

class IDockable ( object ):

    #---------------------------------------------------------------------------
    #  Should the current DockControl be closed before creating the new one:
    #---------------------------------------------------------------------------

    def dockable_should_close ( self ):
        """ Should the current DockControl be closed before creating the new
            one.
        """
        return True

    #---------------------------------------------------------------------------
    #  Returns whether or not it is OK to close the control, and if it is OK,
    #  then it closes the DockControl itself:
    #---------------------------------------------------------------------------

    def dockable_close ( self, dock_control, force ):
        """ Returns whether or not it is OK to close the control.
        """
        return False

    #---------------------------------------------------------------------------
    #  Gets a control that can be docked into a DockWindow:
    #---------------------------------------------------------------------------

    def dockable_get_control ( self, parent ):
        """ Gets a control that can be docked into a DockWindow.
        """
        print("The 'IDockable.dockable_get_control' method must be overridden")
        panel = wx.Panel( parent, -1 )
        panel.SetBackgroundColour( wx.RED )
        return panel

    #---------------------------------------------------------------------------
    #  Allows the object to override the default DockControl settings:
    #---------------------------------------------------------------------------

    def dockable_init_dockcontrol ( self, dock_control ):
        """ Allows the object to override the default DockControl settings.
        """
        pass

    #---------------------------------------------------------------------------
    #  Returns the right-click popup menu for a DockControl (if any):
    #---------------------------------------------------------------------------

    def dockable_menu ( self, dock_control, event ):
        """ Returns the right-click popup menu for a DockControl (if any).
        """
        return None

    #---------------------------------------------------------------------------
    #  Handles the user double-clicking on the DockControl:
    #  A result of False indicates the event was not handled; all other results
    #  indicate that the event was handled successfully.
    #---------------------------------------------------------------------------

    def dockable_dclick ( self, dock_control, event ):
        """ Handles the user double-clicking on the DockControl.
            A result of False indicates the event was not handled; all other
            results indicate that the event was handled successfully.
        """
        return False

    #---------------------------------------------------------------------------
    #  Handles a notebook tab being activated or deactivated:
    #---------------------------------------------------------------------------

    def dockable_tab_activated ( self, dock_control, activated ):
        """ Handles a notebook tab being activated or deactivated.

            'dock_control' is the control being activated or deactivated.

            If 'activated' is True, the control is being activated; otherwise
            the control is being deactivated.
        """
        pass

    #---------------------------------------------------------------------------
    #  Handles the IDockable being bound to a specified DockControl:
    #---------------------------------------------------------------------------

    def dockable_bind ( self, dock_control ):
        """ Handles the dockable being bound to a specified DockControl.
        """
        pass




# Standard library imports.
import unittest

# Local imports.
from traits.api import HasTraits, Int
from pyface.tasks.topological_sort import before_after_sort, \
    topological_sort


class TestItem(HasTraits):

    id = Int
    before = Int
    after = Int

    def __init__(self, id, **traits):
        super(TestItem, self).__init__(id=id, **traits)

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return repr(self.id)


class TopologicalSortTestCase(unittest.TestCase):

    def test_before_after_sort_1(self):
        """ Does the before-after sort work?
        """
        items = [ TestItem(1), TestItem(2), TestItem(3, before=2),
                  TestItem(4, after=1), TestItem(5) ]
        actual = before_after_sort(items)
        desired = [ TestItem(1), TestItem(3), TestItem(4),
                    TestItem(2), TestItem(5) ]
        self.assertEqual(actual, desired)

    def test_before_after_sort_2(self):
        """ Does the before-after sort work when both 'before' and 'after'
            are set?
        """
        items = [ TestItem(1), TestItem(2), TestItem(3),
                  TestItem(4, after=2, before=3) ]
        actual = before_after_sort(items)
        desired = [ TestItem(1), TestItem(2), TestItem(4), TestItem(3) ]
        self.assertEqual(actual, desired)

    def test_before_after_sort_3(self):
        """ Does the degenerate case for the before-after sort work?
        """
        actual = before_after_sort([ TestItem(1) ])
        desired = [ TestItem(1) ]
        self.assertEqual(actual, desired)

    def test_topological_sort_1(self):
        """ Does a basic topological sort work?
        """
        pairs = [ (1,2), (3,5), (4,6), (1,3), (1,4), (1,6), (2,4) ]
        result, has_cycles = topological_sort(pairs)
        self.assertTrue(not has_cycles)
        self.assertEqual(result, [1, 2, 3, 4, 5, 6])

    def test_topological_sort_2(self):
        """ Does another basic topological sort work?
        """
        pairs = [ (1,2), (1,3), (2,4), (3,4), (5,6), (4,5) ]
        result, has_cycles = topological_sort(pairs)
        self.assertTrue(not has_cycles)
        self.assertEqual(result, [1, 2, 3, 4, 5, 6])

    def test_topological_sort_3(self):
        """ Does cycle detection work?
        """
        pairs = [ (1,2), (2,3), (3,1) ]
        result, has_cycles = topological_sort(pairs)
        self.assertTrue(has_cycles)


if __name__ == '__main__':
    unittest.main()



#------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms described
# in the PyQt GPL exception also apply.
#
# Author: Riverbank Computing Limited
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------

# Major package imports.
from pyface.qt import QtGui

# Enthought library imports.
from pyface.action.api import MenuBarManager, StatusBarManager
from pyface.action.api import ToolBarManager
from traits.api import Instance, List, on_trait_change, provides, Unicode

# Local imports.
from pyface.i_application_window import IApplicationWindow, \
     MApplicationWindow
from pyface.image_resource import ImageResource
from .window import Window


@provides(IApplicationWindow)
class ApplicationWindow(MApplicationWindow, Window):
    """ The toolkit specific implementation of an ApplicationWindow.  See the
    IApplicationWindow interface for the API documentation.
    """


    #### 'IApplicationWindow' interface #######################################

    icon = Instance(ImageResource)

    menu_bar_manager = Instance(MenuBarManager)

    status_bar_manager = Instance(StatusBarManager)

    tool_bar_manager = Instance(ToolBarManager)

    tool_bar_managers = List(ToolBarManager)

    #### 'IWindow' interface ##################################################

    title = Unicode("Pyface")

    ###########################################################################
    # Protected 'IApplicationWindow' interface.
    ###########################################################################

    def _create_contents(self, parent):
        panel = QtGui.QWidget(parent)

        palette = QtGui.QPalette(panel.palette())
        palette.setColor(QtGui.QPalette.Window, QtGui.QColor('blue'))
        panel.setPalette(palette)
        panel.setAutoFillBackground(True)

        return panel

    def _create_menu_bar(self, parent):
        if self.menu_bar_manager is not None:
            menu_bar = self.menu_bar_manager.create_menu_bar(parent)
            self.control.setMenuBar(menu_bar)

    def _create_status_bar(self, parent):
        if self.status_bar_manager is not None:
            status_bar = self.status_bar_manager.create_status_bar(parent)

            # QMainWindow automatically makes the status bar visible, but we
            # have delegated this responsibility to the status bar manager.
            self.control.setStatusBar(status_bar)
            status_bar.setVisible(self.status_bar_manager.visible)

    def _create_tool_bar(self, parent):
        tool_bar_managers = self._get_tool_bar_managers()
        visible = self.control.isVisible()
        for tool_bar_manager in tool_bar_managers:
            # Add the tool bar and make sure it is visible.
            tool_bar = tool_bar_manager.create_tool_bar(parent)
            self.control.addToolBar(tool_bar)
            tool_bar.show()

            # Make sure that the tool bar has a name so its state can be saved.
            if len(tool_bar.objectName()) == 0:
                tool_bar.setObjectName(tool_bar_manager.name)

        # Work around bug in Qt on OS X where creating a tool bar with a
        # QMainWindow parent hides the window. See
        # http://bugreports.qt.nokia.com/browse/QTBUG-5069 for more info.
        self.control.setVisible(visible)

    def _set_window_icon(self):
        if self.icon is None:
            icon = ImageResource('application.png')
        else:
            icon = self.icon
        if self.control is not None:
            self.control.setWindowIcon(icon.create_icon())

    ###########################################################################
    # 'Window' interface.
    ###########################################################################

    def _size_default(self):
        """ Trait initialiser. """

        return (800, 600)

    ###########################################################################
    # Protected 'IWidget' interface.
    ###########################################################################

    def _create(self):
        super(ApplicationWindow, self)._create()

        contents = self._create_contents(self.control)
        self.control.setCentralWidget(contents)

        self._create_trim_widgets(self.control)

    def _create_control(self, parent):
        control = super(ApplicationWindow, self)._create_control(parent)
        control.setObjectName('ApplicationWindow')

        control.setAnimated(False)
        control.setDockNestingEnabled(True)

        return control

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_tool_bar_managers(self):
        """ Return all tool bar managers specified for the window. """

        # fixme: V3 remove the old-style single toolbar option!
        if self.tool_bar_manager is not None:
            tool_bar_managers = [self.tool_bar_manager]

        else:
            tool_bar_managers = self.tool_bar_managers

        return tool_bar_managers

    #### Trait change handlers ################################################

    # QMainWindow takes ownership of the menu bar and the status bar upon
    # assignment. For this reason, it is unnecessary to delete the old controls
    # in the following two handlers.

    def _menu_bar_manager_changed(self):
        if self.control is not None:
            self._create_menu_bar(self.control)

    def _status_bar_manager_changed(self):
        if self.control is not None:
            self._create_status_bar(self.control)

    @on_trait_change('tool_bar_manager, tool_bar_managers')
    def _update_tool_bar_managers(self):
        if self.control is not None:
            # Remove the old toolbars.
            for child in self.control.children():
                if isinstance(child, QtGui.QToolBar):
                    self.control.removeToolBar(child)
                    child.deleteLater()

            # Add the new toolbars.
            self._create_tool_bar(self.control)

    def _icon_changed(self):
        self._set_window_icon()



# Enthought library imports.
from pyface.action.api import Action, ActionItem, Group
from traits.api import Any, List, Instance, Property, Unicode, on_trait_change

# Local imports.
from pyface.tasks.task import Task
from pyface.tasks.task_window import TaskWindow


class TaskToggleAction(Action):
    """ An action for activating a task.
    """

    #### 'Action' interface ###################################################

    name = Property(Unicode, depends_on='task.name')
    style = 'toggle'
    tooltip = Property(Unicode, depends_on='name')

    #### 'TaskActivateAction' interface #######################################
    
    task = Instance(Task)
    
    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def destroy(self):
        super(TaskToggleAction, self).destroy()

        # Make sure that we are not listening to changes in the task anymore
        # In traits style, we will set the basic object to None and have the
        # listener check that if it is still there.
        self.task = None

    def perform(self, event=None):
        window = self.task.window
        window.activate_task(self.task)

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_name(self):
        if self.task is None:
            return 'UNDEFINED'
        return self.task.name

    def _get_tooltip(self):
        return 'Switch to the %s task.' % self.name

    @on_trait_change('task.window.active_task')
    def _update_checked(self):
        if self.task:
            window = self.task.window
            self.checked = (window is not None
                            and window.active_task == self.task)


class TaskToggleGroup(Group):
    """ A menu for changing the active task in a task window.
    """

    #### 'ActionManager' interface ############################################

    id = 'TaskToggleGroup'
    items = List

    #### 'TaskChangeMenuManager' interface ####################################

    # The ActionManager to which the group belongs.
    manager = Any

    # The window that contains the group.
    window = Instance(TaskWindow)
        
    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_items(self):
        items = []
        for task in self.window.tasks:
            action = TaskToggleAction(task=task)
            items.append(ActionItem(action=action))
        return items

    def _rebuild(self):
        # Clear out the old group, then build the new one.
        self.destroy()
        self.items = self._get_items()

        # Inform our manager that it needs to be rebuilt.
        self.manager.changed = True
        
    #### Trait initializers ###################################################

    def _items_default(self):
        self.window.on_trait_change(self._rebuild, 'tasks[]')
        return self._get_items()

    def _manager_default(self):
        manager = self
        while isinstance(manager, Group):
            manager = manager.parent
        return manager
    
    def _window_default(self):
        return self.manager.controller.task.window



#-------------------------------------------------------------------------------
#
#  Copyright (c) 2006, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: Prabhu Ramachandran <prabhu@aero.iitb.ac.in>
#
#-------------------------------------------------------------------------------

"""A `wx.Timer` subclass that invokes a specified callback periodically.
"""

# Standard library imports.
import wx

######################################################################
# `Timer` class.
class Timer(wx.Timer):
    """Simple subclass of wx.Timer that allows the user to have a
    function called periodically.

    Any exceptions raised in the callable are caught.  If
    `StopIteration` is raised the timer stops.  If other exceptions are
    encountered the timer is stopped and the exception re-raised.
    """

    def __init__(self, millisecs, callable, *args, **kw_args):
        """Initialize instance to invoke the given `callable` with
        given arguments and keyword args after every `millisecs`
        (milliseconds).
        """
        wx.Timer.__init__(self, id=wx.NewId())
        self.callable = callable
        self.args = args
        self.kw_args = kw_args
        self.Start(millisecs)

    def Notify(self):
        """Overridden to call the given callable.  Exceptions raised
        in the callable are caught.  If `StopIteration` is raised the
        timer stops.  If other exceptions are encountered the timer is
        stopped and the exception re-raised.
        """
        try:
            self.callable(*self.args, **self.kw_args)
            wx.GetApp().Yield(True)
        except StopIteration:
            self.Stop()
        except:
            self.Stop()
            raise




""" Abstract base class for all workbench actions. """


# Enthought library imports.
from pyface.workbench.api import WorkbenchWindow
from pyface.action.api import Action
from traits.api import Instance


class WorkbenchAction(Action):
    """ Abstract base class for all workbench actions. """

    #### 'WorkbenchAction' interface ##########################################

    # The workbench window that the action is in.
    #
    # This is set by the framework.
    window = Instance(WorkbenchWindow)

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" The interface for a dialog that allows the user to open/save files etc. """


# Standard library imports.
import sys

# Enthought library imports.
from traits.api import Enum, Unicode, Int

# Local imports.
from pyface.i_dialog import IDialog


class IFileDialog(IDialog):
    """ The interface for a dialog that allows the user to open/save files etc.
    """

    #### 'IFileDialog' interface ##############################################

    #: The 'action' that the user is peforming on the directory.
    action = Enum('open', 'save as')

    #: The default directory.
    default_directory = Unicode

    #: The default filename.
    default_filename = Unicode

    #: The default path (directory and filename) of the chosen file.  This is
    #: only used when the *default_directory* and *default_filename* are not set
    #: and is equivalent to setting both.
    default_path = Unicode

    #: The directory containing the chosen file.
    directory = Unicode

    #: The name of the chosen file.
    filename = Unicode

    #: The path (directory and filename) of the chosen file.
    path = Unicode

    #: The wildcard used to restrict the set of files.
    wildcard = Unicode

    #: The index of the selected wildcard.
    wildcard_index = Int(0)


class MFileDialog(object):
    """ The mixin class that contains common code for toolkit specific
    implementations of the IFileDialog interface.

    Implements: create_wildcard()
    """

    # FIXME v3: These are referenced elsewhere so maybe should be part of the
    # interface. The format is toolkit specific and so shouldn't be exposed.
    # The create_wildcard() class method (why isn't it a static method?) should
    # be the basis for this - but nothing seems to use it.  For now the PyQt
    # implementation will convert the wx format to its own.  Actually we need
    # the format to be portable between toolkits - so stick with PyQt
    # supporting the wx format or invent a data structure.

    #: A file dialog wildcard for Python files.
    WILDCARD_PY = "Python files (*.py)|*.py|"

    #: A file dialog wildcard for text files.
    WILDCARD_TXT = "Text files (*.txt)|*.txt|"

    #: A file dialog wildcard for all files.
    if sys.platform == 'win32':
        WILDCARD_ALL = "All files (*.*)|*.*|"
    else:
        WILDCARD_ALL = "All files (*)|*"

    #: A file dialog wildcard for Zip archives.
    WILDCARD_ZIP = "Zip files (*.zip)|*.zip|"

    ###########################################################################
    # 'MFileDialog' *CLASS* interface.
    ###########################################################################

    @classmethod
    def create_wildcard(cls, description, extension):
        """ Creates a wildcard for a given extension.

        Parameters
        ----------
        description : str
            A human-readable description of the pattern.
        extenstion : list
            The wildcard patterns for the extension.
        """

        if isinstance(extension, str):
            pattern = extension

        else:
            pattern = ';'.join(extension)

        return "%s (%s)|%s|" % (description, pattern, pattern)



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" A group of action manager items. """


# Enthought library imports.
from traits.api import Any, Bool, HasTraits, Instance, List, Property
from traits.api import Str
from traits.trait_base import user_name_for

# Local imports.
from pyface.action.action import Action
from pyface.action.action_item import ActionItem
from pyface.action.action_manager_item import ActionManagerItem


class Group(HasTraits):
    """ A group of action manager items.

    By default, a group declares itself as requiring a separator when it is
    visualized, but this can be changed by setting its 'separator' trait to
    False.

    """

    #### 'Group' interface ####

    #: Is the group enabled?
    enabled = Bool(True)

    #: Is the group visible?
    visible = Bool(True)

    #: The group's unique identifier (only needs to be unique within the action
    #: manager that the group belongs to).
    id = Str

    #: All of the items in the group.
    items = Property

    #: The action manager that the group belongs to.
    parent = Any#Instance('pyface.action.ActionManager')

    #: Does this group require a separator when it is visualized?
    separator = Bool(True)

    #### Private interface ####

    #: All of the items in the group.
    _items = List#(ActionManagerItem)

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, *items,  **traits):
        """ Creates a new menu manager.

        Parameters
        ----------
        items : collection of ActionManagerItems
            Items to add to the group.
        """
        # Base class constructor.
        super(Group, self).__init__(**traits)

        # Add any specified items.
        for item in items:
            self.append(item)

    ###########################################################################
    # 'Group' interface.
    ###########################################################################

    #### Trait Properties #####################################################

    def _get_items(self):
        return self._items[:]

    #### Trait change handlers ################################################

    def _enabled_changed(self, trait_name, old, new):
        for item in self.items:
            item.enabled = new

    #### Methods ##############################################################

    def append(self, item):
        """ Appends an item to the group.

        Parameters
        ----------
        item : ActionManagerItem, Action or callable
            The item to append.

        Returns
        -------
        item : ActionManagerItem
            The actually inserted item.

        Notes
        -----
        If the item is an ActionManagerItem instance it is simply appended.
        If the item is an Action instance, an ActionItem is created for the
        action, and that is appended.  If the item is a callable, then an
        Action is created for the callable, and then that is handled as above.
        """
        return self.insert(len(self._items), item)

    def clear(self):
        """ Remove all items from the group. """
        self._items = []

    def destroy(self):
        """ Called when the manager is no longer required.

        By default this method simply calls 'destroy' on all items in the
        group.
        """
        for item in self.items:
            item.destroy()

    def insert(self, index, item):
        """ Inserts an item into the group at the specified index.

        Parameters
        ----------
        index : int
            The position to insert the item.
        item : ActionManagerItem, Action or callable
            The item to insert.

        Returns
        -------
        item : ActionManagerItem
            The actually inserted item.

        Notes
        -----

        If the item is an ActionManagerItem instance it is simply inserted.
        If the item is an Action instance, an ActionItem is created for the
        action, and that is inserted.  If the item is a callable, then an
        Action is created for the callable, and then that is handled as above.
        """
        if isinstance(item, Action):
            item = ActionItem(action=item)
        elif callable(item):
            name = user_name_for(item.__name__)
            item = ActionItem(action=Action(name=name, on_perform=item))

        item.parent = self
        self._items.insert(index, item)

        return item

    def remove(self, item):
        """ Removes an item from the group.

        Parameters
        ----------
        item : ActionManagerItem
            The item to remove.
        """
        self._items.remove(item)
        item.parent = None

    def insert_before(self, before, item):
        """ Inserts an item into the group before the specified item.

        Parameters
        ----------
        before : ActionManagerItem
            The item to insert before.
        item : ActionManagerItem, Action or callable
            The item to insert.

        Returns
        -------
        index, item : int, ActionManagerItem
            The position inserted, and the item actually inserted.

        Notes
        -----

        If the item is an ActionManagerItem instance it is simply inserted.
        If the item is an Action instance, an ActionItem is created for the
        action, and that is inserted.  If the item is a callable, then an
        Action is created for the callable, and then that is handled as above.
        """
        index = self._items.index(before)
        self.insert(index, item)
        return (index, item)

    def insert_after(self, after, item):
        """ Inserts an item into the group after the specified item.

        Parameters
        ----------
        before : ActionManagerItem
            The item to insert after.
        item : ActionManagerItem, Action or callable
            The item to insert.

        Returns
        -------
        index, item : int, ActionManagerItem
            The position inserted, and the item actually inserted.

        Notes
        -----

        If the item is an ActionManagerItem instance it is simply inserted.
        If the item is an Action instance, an ActionItem is created for the
        action, and that is inserted.  If the item is a callable, then an
        Action is created for the callable, and then that is handled as above.
        """
        index = self._items.index(after)
        self.insert(index + 1, item)
        return (index, item)

    def find(self, id):
        """ Find the item with the specified id.

        Parameters
        ----------
        id : str
            The id of the item

        Returns
        -------
        item : ActionManagerItem
            The item with the specified Id, or None if no such item exists.
        """
        for item in self._items:
            if item.id == id:
                return item
        else:
            return None


class Separator(Group):
    """ A convenience class.

    This is only used in 'cheap and cheerful' applications that create menus
    like::

        file_menu = MenuManager(
            CopyAction(),
            Separator(),
            ExitAction()
        )

    Hopefully, 'Separator' is more readable than 'Group'...
    """
    pass





from traits.testing.unittest_tools import unittest, UnittestTools

from ..toolkit import toolkit_object
from ..widget import Widget

GuiTestAssistant = toolkit_object('util.gui_test_assistant:GuiTestAssistant')
no_gui_test_assistant = (GuiTestAssistant.__name__ == 'Unimplemented')


class ConcreteWidget(Widget):

    def _create_control(self, parent):
        if toolkit_object.toolkit == 'wx':
            import wx
            control = wx.Window(parent)
            control.Enable(self.enabled)
            control.Show(self.visible)
        elif toolkit_object.toolkit == 'qt4':
            from pyface.qt import QtGui
            control = QtGui.QWidget(parent)
            control.setEnabled(self.enabled)
            control.setVisible(self.visible)
        else:
            control = None
        return control


class TestWidget(unittest.TestCase, UnittestTools):

    def setUp(self):
        self.widget = Widget()

    def tearDown(self):
        del self.widget

    def test_defaults(self):
        self.assertTrue(self.widget.enabled)
        self.assertTrue(self.widget.visible)

    def test_create(self):
        # create is not Implemented
        with self.assertRaises(NotImplementedError):
            self.widget._create()

    def test_destroy(self):
        # test that destroy works even when no control
        self.widget.destroy()

    def test_show(self):
        with self.assertTraitChanges(self.widget, 'visible', count=1):
            self.widget.show(False)

        self.assertFalse(self.widget.visible)

    def test_visible(self):
        with self.assertTraitChanges(self.widget, 'visible', count=1):
            self.widget.visible = False

        self.assertFalse(self.widget.visible)

    def test_enable(self):
        with self.assertTraitChanges(self.widget, 'enabled', count=1):
            self.widget.enable(False)

        self.assertFalse(self.widget.enabled)

    def test_enabled(self):
        with self.assertTraitChanges(self.widget, 'enabled', count=1):
            self.widget.enabled = False

        self.assertFalse(self.widget.enabled)


@unittest.skipIf(no_gui_test_assistant, 'No GuiTestAssistant')
class TestConcreteWidget(unittest.TestCase, GuiTestAssistant):

    def setUp(self):
        GuiTestAssistant.setUp(self)
        self.widget = ConcreteWidget()

    def tearDown(self):
        if self.widget.control is not None:
            with self.delete_widget(self.widget.control):
                self.widget.destroy()
        del self.widget
        GuiTestAssistant.tearDown(self)

    def test_lifecycle(self):
        self.widget._create()
        self.event_loop()
        self.widget.destroy()
        self.event_loop()

    def test_initialize(self):
        self.widget.visible = False
        self.widget.enabled = False
        self.widget._create()
        self.event_loop()

        self.assertFalse(self.widget.control.isVisible())
        self.assertFalse(self.widget.control.isEnabled())

    def test_show(self):
        self.widget._create()
        self.event_loop()

        with self.assertTraitChanges(self.widget, 'visible', count=1):
            self.widget.show(False)
            self.event_loop()

        self.assertFalse(self.widget.control.isVisible())

    def test_visible(self):
        self.widget._create()
        self.event_loop()

        with self.assertTraitChanges(self.widget, 'visible', count=1):
            self.widget.visible = False
            self.event_loop()

        self.assertFalse(self.widget.control.isVisible())

    def test_enable(self):
        self.widget._create()
        self.event_loop()

        with self.assertTraitChanges(self.widget, 'enabled', count=1):
            self.widget.enable(False)
            self.event_loop()

        self.assertFalse(self.widget.control.isEnabled())

    def test_enabled(self):
        self.widget._create()
        self.event_loop()

        with self.assertTraitChanges(self.widget, 'enabled', count=1):
            self.widget.enabled = False
            self.event_loop()

        self.assertFalse(self.widget.control.isEnabled())



#------------------------------------------------------------------------------
# Copyright (c) 2005-2013, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" Reusable MVC-based components for Traits-based applications.
    Part of the TraitsGUI project of the Enthought Tool Suite.
"""

try:
    from pyface._version import full_version as __version__
except ImportError:
    __version__ = 'not-built'


__requires__ = ['traits']
__extras_require__ = {
    'wx': ['wxpython>=2.8.10', 'numpy'],
    'pyqt': ['pyqt>=4.10', 'pygments'],
    'pyqt5': ['pyqt>=5', 'pygments'],
    'pyside': ['pyside>=1.2', 'pygments'],
}



#------------------------------------------------------------------------------
# Copyright (c) 2006, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" A renderer which displays a checked-box for a True value and an unchecked
    box for a false value. """

# Enthought-library imports
from pyface.image_resource import ImageResource

# local imports
from .mapped_grid_cell_image_renderer import MappedGridCellImageRenderer

checked_image_map = { True: ImageResource('checked'),
                      False: ImageResource('unchecked'),
                    }

class CheckboxImageRenderer(MappedGridCellImageRenderer):

    def __init__(self, display_text = False):

        text_map = None
        if display_text:
            text_map = { True: 'True', False: 'False' }

        # Base-class constructor
        super(CheckboxImageRenderer, self).__init__(checked_image_map,
                                                    text_map)

        return

#### EOF ######################################################################



#-------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   11/04/2005
#
#-------------------------------------------------------------------------------

""" Defines the DockWindowShell class used to house drag and drag DockWindow
    items that are dropped on the desktop or on the DockWindowShell window.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import wx

# Fixme: Hack to force 'image_slice' to be added via Category to Theme class:
import traitsui.wx

from traits.api \
    import HasPrivateTraits, Instance

from traitsui.api \
    import View, Group

from pyface.api import SystemMetrics

from pyface.image_resource \
    import ImageResource

from .dock_window \
    import DockWindow

from .dock_sizer \
    import DockSizer, DockSection, DockRegion, DockControl, DOCK_RIGHT

#-------------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------------

# DockWindowShell frame icon:
FrameIcon = ImageResource( 'shell.ico' )

#-------------------------------------------------------------------------------
#  'DockWindowShell' class:
#-------------------------------------------------------------------------------

class DockWindowShell ( HasPrivateTraits ):

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # The wx.Frame window which is the actual shell:
    control = Instance( wx.Frame )

    #---------------------------------------------------------------------------
    #  Initializes the object:
    #---------------------------------------------------------------------------

    def __init__ ( self, dock_control, use_mouse = False, **traits ):
        super( DockWindowShell, self ).__init__( **traits )

        old_control = dock_control.control
        parent      = wx.GetTopLevelParent( old_control )
        while True:
            next_parent = parent.GetParent()
            if next_parent is None:
                break
            parent = next_parent

        self.control = shell = wx.Frame( parent, -1, dock_control.name,
                                         style = wx.DEFAULT_FRAME_STYLE   |
                                                 wx.FRAME_FLOAT_ON_PARENT |
                                                 wx.FRAME_NO_TASKBAR )
        shell.SetIcon( FrameIcon.create_icon() )
        shell.SetBackgroundColour( SystemMetrics().dialog_background_color )
        wx.EVT_CLOSE( shell, self._on_close )

        theme = dock_control.theme
        self._dock_window = dw = DockWindow( shell, auto_close = True,
                                                    theme      = theme ).set(
                                                    style      = 'tab' )
        sizer = wx.BoxSizer( wx.VERTICAL )
        sizer.Add( dw.control, 1, wx.EXPAND )
        shell.SetSizer( sizer )

        if use_mouse:
            x, y = wx.GetMousePosition()
        else:
            x, y = old_control.GetPositionTuple()
            x, y = old_control.GetParent().ClientToScreenXY( x, y )

        dx, dy = old_control.GetSize()
        tis    = theme.tab.image_slice
        tc     = theme.tab.content
        tdy    = theme.tab_active.image_slice.dy
        dx    += (tis.xleft + tc.left + tis.xright  + tc.right)
        dy    += (tis.xtop  + tc.top  + tis.xbottom + tc.bottom + tdy)

        self.add_control( dock_control )

        # Set the correct window size and position, accounting for the tab size
        # and window borders:
        shell.SetDimensions( x, y, dx, dy )
        cdx, cdy = shell.GetClientSizeTuple()
        ex_dx    = dx - cdx
        ex_dy    = dy - cdy
        shell.SetDimensions( x - (ex_dx / 2) - tis.xleft - tc.left,
                             y - ex_dy + (ex_dx / 2) - tdy - tis.xtop - tc.top,
                             dx + ex_dx, dy + ex_dy )
        shell.Show()

    #---------------------------------------------------------------------------
    #  Adds a new DockControl to the shell window:
    #---------------------------------------------------------------------------

    def add_control ( self, dock_control ):
        """ Adds a new DockControl to the shell window.
        """
        dw       = self._dock_window.control
        dockable = dock_control.dockable

        # If the current DockControl should be closed, then do it:
        close = dockable.dockable_should_close()
        if close:
            dock_control.close( force = True )

        # Create the new control:
        control = dockable.dockable_get_control( dw )

        # If the DockControl was closed, then reset it to point to the new
        # control:
        if close:
            dock_control.set( control = control, style = 'tab' )
        else:
            # Create a DockControl to describe the new control:
            dock_control = DockControl( control   = control,
                                        name      = dock_control.name,
                                        export    = dock_control.export,
                                        style     = 'tab',
                                        image     = dock_control.image,
                                        closeable = True )

        # Finish initializing the DockControl:
        dockable.dockable_init_dockcontrol( dock_control )

        # Get the current DockSizer:
        sizer = dw.GetSizer()
        if sizer is None:
            # Create the initial sizer:
            dw.SetSizer( DockSizer( DockSection( contents = [ DockRegion(
                                    contents = [ dock_control ] ) ] ) ) )
        else:
            # Sizer exists already, try to add the DockControl as a new
            # notebook tab. If the user has reorganized the layout, then just
            # dock it on the right side somewhere:
            section = sizer.GetContents()
            region  = section.contents[0]
            if isinstance( region, DockRegion ):
                region.add( dock_control )
            else:
                section.add( dock_control, region, DOCK_RIGHT )

            # Force the control to update:
            dw.Layout()
            dw.Refresh()

    #---------------------------------------------------------------------------
    #  Handles the user attempting to close the window:
    #---------------------------------------------------------------------------

    def _on_close ( self, event ):
        """ Handles the user attempting to close the window.
        """
        window  = self._dock_window.control
        section = window.GetSizer().GetContents()
        n       = len( section.contents )

        # Try to close each individual control:
        for control in section.get_controls():
            control.close( layout = False )

        # If some, but not all, were closed, make sure the window gets updated:
        if 0 < len( section.contents ) < n:
            window.Layout()
            window.Refresh()




#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: Enthought, Inc.
#
#------------------------------------------------------------------------------

""" Enthought pyface package component
"""

# Major package imports.
import wx

# Enthought library imports.
from traits.api import Enum, provides, Unicode

# Local imports.
from pyface.i_message_dialog import IMessageDialog, MMessageDialog
from .dialog import Dialog


# Map the ETS severity to the corresponding wx standard icon.
_SEVERITY_TO_ICON_MAP = {
    'information':  wx.ICON_INFORMATION,
    'warning':      wx.ICON_WARNING,
    'error':        wx.ICON_ERROR
}


@provides(IMessageDialog)
class MessageDialog(MMessageDialog, Dialog):
    """ The toolkit specific implementation of a MessageDialog.  See the
    IMessageDialog interface for the API documentation.
    """


    #### 'IMessageDialog' interface ###########################################

    message = Unicode

    informative = Unicode

    detail = Unicode

    severity = Enum('information', 'warning', 'error')

    ###########################################################################
    # Protected 'IDialog' interface.
    ###########################################################################

    def _create_contents(self, parent):
        # In wx this is a canned dialog.
        pass

    ###########################################################################
    # Protected 'IWidget' interface.
    ###########################################################################

    def _create_control(self, parent):
        # The message.
        if self.informative:
            message = self.message + '\n\n' + self.informative
        else:
            message = self.message

        style = _SEVERITY_TO_ICON_MAP[self.severity] | wx.OK | wx.STAY_ON_TOP
        if self.resizeable:
            style |= wx.RESIZE_BORDER

        dlg = wx.MessageDialog(parent, message, self.title, style,
                               self.position)
        if self.size != (-1, -1):
            dlg.SetSize(self.size)

        return dlg




import mock

from traits.testing.unittest_tools import unittest

from pyface.ui.qt4.workbench.split_tab_widget import SplitTabWidget
from pyface.ui.qt4.workbench.workbench_window_layout import \
    WorkbenchWindowLayout


class TestWorkbenchWindowLayout(unittest.TestCase):
    def test_change_of_active_qt_editor(self):
        # Test error condition for enthought/mayavi#321

        mock_split_tab_widget = mock.Mock(spec=SplitTabWidget)

        layout = WorkbenchWindowLayout(_qt4_editor_area=mock_split_tab_widget)

        # This should not throw
        layout._qt4_active_editor_changed(None, None)
        self.assertEqual(mock_split_tab_widget.setTabTextColor.called, False)

        mock_active_editor = mock.Mock()
        layout._qt4_active_editor_changed(None, mock_active_editor)

        self.assertEqual(mock_split_tab_widget.setTabTextColor.called, True)





from traits.testing.unittest_tools import UnittestTools, unittest

from ...image_cache import ImageCache
from ...window import Window
from ..action import Action
from ..action_controller import ActionController
from ..action_item import ActionItem
from ..menu_manager import MenuManager
from ..menu_bar_manager import MenuBarManager
from ..tool_bar_manager import ToolBarManager


class FalseActionController(ActionController):

    def can_add_to_menu(self, action):
        """ Returns True if the action can be added to a menu/menubar. """

        return False

    def can_add_to_toolbar(self, action):
        """ Returns True if the action can be added to a toolbar. """

        return False


class TestActionItem(unittest.TestCase, UnittestTools):

    def setUp(self):
        # test whether function is called by updating list
        # XXX should really use mock
        self.memo = []

        def perform():
            self.memo.append('called')

        self.action = Action(name='Test', on_perform=perform)

    def test_default_id(self):
        action_item = ActionItem(action=self.action)
        self.assertEqual(action_item.id, 'Test')

    def test_enabled_changed(self):
        # XXX these are only one-way changes, which seems wrong.
        action_item = ActionItem(action=self.action)
        with self.assertTraitChanges(self.action, 'enabled', count=1):
            action_item.enabled = False
        self.assertFalse(self.action.enabled)
        with self.assertTraitChanges(self.action, 'enabled', count=1):
            action_item.enabled = True
        self.assertTrue(self.action.enabled)

    def test_visible_changed(self):
        # XXX these are only one-way changes, which seems wrong.
        action_item = ActionItem(action=self.action)
        with self.assertTraitChanges(self.action, 'visible', count=1):
            action_item.visible = False
        self.assertFalse(self.action.visible)
        with self.assertTraitChanges(self.action, 'visible', count=1):
            action_item.visible = True
        self.assertTrue(self.action.visible)

    def test_destroy(self):
        action_item = ActionItem(action=self.action)
        # XXX test that it calls action.destroy
        action_item.destroy()

    def test_add_to_menu(self):
        window = Window()
        window.open()
        action_item = ActionItem(action=self.action)
        menu_bar_manager = MenuBarManager()
        menu_manager = MenuManager(name='Test')
        menu_bar = menu_bar_manager.create_menu_bar(window.control)
        menu = menu_manager.create_menu(menu_bar)
        action_item.add_to_menu(window.control, menu, None)
        window.close()

    def test_add_to_menu_controller(self):
        window = Window()
        window.open()
        action_item = ActionItem(action=self.action)
        menu_bar_manager = MenuBarManager()
        menu_manager = MenuManager(name='Test')
        menu_bar = menu_bar_manager.create_menu_bar(window.control)
        menu = menu_manager.create_menu(menu_bar)
        controller = ActionController()
        action_item.add_to_menu(window.control, menu, controller)
        window.close()

    def test_add_to_menu_controller_false(self):
        window = Window()
        window.open()
        action_item = ActionItem(action=self.action)
        menu_bar_manager = MenuBarManager()
        menu_manager = MenuManager(name='Test')
        menu_bar = menu_bar_manager.create_menu_bar(window.control)
        menu = menu_manager.create_menu(menu_bar)
        controller = FalseActionController()
        action_item.add_to_menu(window.control, menu, controller)
        window.close()

    def test_add_to_toolbar(self):
        window = Window()
        window.open()
        action_item = ActionItem(action=self.action)
        toolbar_manager = ToolBarManager(name='Test')
        image_cache = ImageCache(height=32, width=32)
        menu = toolbar_manager.create_tool_bar(window.control)
        action_item.add_to_toolbar(window.control, menu, image_cache, None, True)
        window.close()

    def test_add_to_toolbar_no_label(self):
        window = Window()
        window.open()
        action_item = ActionItem(action=self.action)
        toolbar_manager = ToolBarManager(name='Test')
        image_cache = ImageCache(height=32, width=32)
        menu = toolbar_manager.create_tool_bar(window.control)
        action_item.add_to_toolbar(window.control, menu, image_cache, None, False)
        window.close()

    def test_add_to_toolbar_controller(self):
        window = Window()
        window.open()
        action_item = ActionItem(action=self.action)
        toolbar_manager = ToolBarManager(name='Test')
        image_cache = ImageCache(height=32, width=32)
        menu = toolbar_manager.create_tool_bar(window.control)
        controller = ActionController()
        action_item.add_to_toolbar(window.control, menu, image_cache,
                                   controller, True)
        window.close()

    def test_add_to_toolbar_controller_false(self):
        window = Window()
        window.open()
        action_item = ActionItem(action=self.action)
        toolbar_manager = ToolBarManager(name='Test')
        image_cache = ImageCache(height=32, width=32)
        menu = toolbar_manager.create_tool_bar(window.control)
        controller = FalseActionController()
        action_item.add_to_toolbar(window.control, menu, image_cache,
                                   controller, True)
        window.close()



import pkg_resources

from traits.testing.unittest_tools import unittest

import pyface.toolkit


class TestToolkit(unittest.TestCase):

    def test_missing_import(self):
        # test that we get an undefined object if no toolkit implementation
        cls = pyface.toolkit.toolkit_object('tests:Missing')
        with self.assertRaises(NotImplementedError):
            obj = cls()

    def test_bad_import(self):
        # test that we don't filter unrelated import errors
        with self.assertRaises(ImportError):
            cls = pyface.toolkit.toolkit_object('tests.bad_import:Missing')

    def test_core_plugins(self):
        # test that we can see appropriate core entrypoints
        plugins = set(entry_point.name for entry_point in
                      pkg_resources.iter_entry_points('pyface.toolkits'))

        self.assertLessEqual({'qt4', 'wx', 'qt', 'null'}, plugins)

    def test_toolkit_object(self):
        # test that the Toolkit class works as expected
        # note that if this fails many other things will too
        from pyface.tests.test_new_toolkit.init import toolkit_object
        from pyface.tests.test_new_toolkit.widget import Widget as TestWidget

        Widget = toolkit_object('widget:Widget')

        self.assertEqual(Widget, TestWidget)

    def test_toolkit_object_overriden(self):
        # test that the Toolkit class search paths can be overridden
        from pyface.tests.test_new_toolkit.widget import Widget as TestWidget

        toolkit_object = pyface.toolkit.toolkit_object

        old_packages = toolkit_object.packages
        toolkit_object.packages = ['pyface.tests.test_new_toolkit'] + old_packages
        try:
            Widget = toolkit_object('widget:Widget')
            self.assertEqual(Widget, TestWidget)
        finally:
            toolkit_object.packages = old_packages

    def test_toolkit_object_not_overriden(self):
        # test that the Toolkit class works when object not overridden
        toolkit_object = pyface.toolkit.toolkit_object
        TestWindow = toolkit_object('window:Window')

        old_packages = toolkit_object.packages
        toolkit_object.packages = ['pyface.tests.test_new_toolkit'] + old_packages
        try:
            Window = toolkit_object('window:Window')
            self.assertEqual(Window, TestWindow)
        finally:
            toolkit_object.packages = old_packages



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2005-2006 by Enthought, Inc.
#  All rights reserved.
#
#  Author: David C. Morrill <dmorrill@enthought.com>
#
#-----------------------------------------------------------------------------
""" An action that creates a new (and empty) user perspective. """


# Local imports.
from .user_perspective_name import UserPerspectiveName
from .workbench_action import WorkbenchAction


class NewUserPerspectiveAction(WorkbenchAction):
    """ An action that creates a new (and empty) user perspective. """

    #### 'Action' interface ###################################################

    # The action's unique identifier.
    id = 'pyface.workbench.action.new_user_perspective_action'

    # The action's name.
    name = 'New Perspective...'

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Peform the action. """

        window  = event.window
        manager = window.workbench.user_perspective_manager

        # Get the details of the new perspective.
        upn = UserPerspectiveName(name='User Perspective %d' % manager.next_id)
        if upn.edit_traits(view='new_view').result:
            # Create a new (and empty) user perspective.
            perspective = manager.create_perspective(
                upn.name.strip(), upn.show_editor_area
            )

            # Add it to the window...
            window.perspectives.append(perspective)

            # ... and make it the active perspective.
            window.active_perspective = perspective

        return

#### EOF #####################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" Abstract base class for all viewer sorters. """


# Enthought library imports.
from traits.api import HasTraits


class ViewerSorter(HasTraits):
    """ Abstract base class for all viewer sorters. """

    ###########################################################################
    # 'ViewerSorter' interface.
    ###########################################################################

    def sort(self, viewer, parent, elements):
        """ Sorts a list of elements IN PLACE.

        'viewer'   is the viewer that we are sorting elements for.
        'parent'   is the parent element.
        'elements' is the list of elements to sort.

        Returns the list that was sorted IN PLACE (for convenience).

        """

        # This creates a comparison function with the names 'viewer' and
        # 'parent' bound to the corresponding arguments to this method.
        def comparator(element_a, element_b):
            """ Comparator. """

            return self.compare(viewer, parent, element_a, element_b)

        elements.sort(comparator)

        return elements

    def compare(self, viewer, parent, element_a, element_b):
        """ Returns the result of comparing two elements.

        'viewer'    is the viewer that we are sorting elements for.
        'parent'    is the parent element.
        'element_a' is the the first element to compare.
        'element_b' is the the second element to compare.

        """

        # Get the category for each element.
        category_a = self.category(viewer, parent, element_a)
        category_b = self.category(viewer, parent, element_b)

        # If they are not in the same category then return the result of
        # comparing the categories.
        if category_a != category_b:
            result = cmp(category_a, category_b)

        else:
            # Get the label text for each element.
            #
            # fixme: This is a hack until we decide whethwe we like the
            # JFace(ish) or Swing(ish) models!
            if hasattr(viewer, 'label_provider'):
              label_a = viewer.label_provider.get_text(viewer, element_a)
              label_b = viewer.label_provider.get_text(viewer, element_b)

            else:
                label_a = viewer.node_model.get_text(viewer, element_a)
                label_b = viewer.node_model.get_text(viewer, element_b)

            # Compare the label text.
            result = cmp(label_a, label_b)

        return result

    def category(self, viewer, parent, element):
        """ Returns the category (an integer) for an element.

        'parent'   is the parent element.
        'elements' is the element to return the category for.

        Categories are used to sort elements into bins.  The bins are
        arranged in ascending numeric order.  The elements within a bin
        are arranged as dictated by the sorter's 'compare' method.

        By default all elements are given the same category (0).

        """

        return 0

    def is_sorter_trait(self, element, trait_name):
        """ Is the sorter affected by changes to an element's trait?

        'element'    is the element.
        'trait_name' is the name of the trait.

        Returns True if the sorter would be affected by changes to the trait
        named 'trait_name' on the specified element.

        By default we return False.

        """

        return False

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# Copyright (c) 2017, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license.  However, when used with the GPL version of PyQt the additional
# terms described in the PyQt GPL exception also apply.
#
# Author: Riverbank Computing Limited and Enthought developers
#------------------------------------------------------------------------------


import sys

from pyface.qt import QtCore, QtGui, qt_api
from pyface.base_toolkit import Toolkit

if qt_api == 'pyqt':
    # Check the version numbers are late enough.
    if QtCore.QT_VERSION < 0x040200:
        raise RuntimeError(
            "Need Qt v4.2 or higher, but got v%s" % QtCore.QT_VERSION_STR
        )

    if QtCore.PYQT_VERSION < 0x040100:
        raise RuntimeError(
            "Need PyQt v4.1 or higher, but got v%s" % QtCore.PYQT_VERSION_STR
        )

# It's possible that it has already been initialised.
_app = QtGui.QApplication.instance()

if _app is None:
    _app = QtGui.QApplication(sys.argv)


# create the toolkit object
toolkit_object = Toolkit('pyface', 'qt4', 'pyface.ui.qt4')



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" Abstract base class for label providers. """


# Enthought library imports.
from traits.api import HasTraits


class LabelProvider(HasTraits):
    """ Abstract base class for label providers.

    By default an element has no label image, and 'str' is used to generate its
    label text.

    """

    ###########################################################################
    # 'LabelProvider' interface.
    ###########################################################################

    def get_image(self, viewer, element):
        """ Returns the label image for an element. """

        return None

    def get_text(self, viewer, element):
        """ Returns the label text for an element. """

        return str(element)

    def set_text(self, tree, element, text):
        """ Sets the text representation of a node.

        Returns True if setting the text succeeded, otherwise False.

        """

        return True

    def is_editable(self, viewer, element):
        """ Can the label text be changed via the viewer? """

        return False

#### EOF ######################################################################



# Enthought library imports.
from traits.api import Either, List, Str, Tuple, Enum

# Local imports.
from pyface.tasks.task_layout import LayoutContainer, TaskLayout


class TaskWindowLayout(LayoutContainer):
    """ The layout of a TaskWindow.
    """

    # The ID of the active task. If unspecified, the first task will be active.
    active_task = Str

    # The tasks contained in the window. If an ID is specified, the task will
    # use its default layout. Otherwise, it will use the specified TaskLayout.
    items = List(Either(Str, TaskLayout), pretty_skip=True)

    # The position of the window.
    position = Tuple(-1, -1)

    # The size of the window.
    size = Tuple(800, 600)

    size_state = Enum('normal', 'maximized')

    def get_active_task(self):
        """ Returns the ID of the active task in the layout, or None if there is
            no active task.
        """
        if self.active_task:
            return self.active_task
        elif self.items:
            first = self.items[0]
            return first if isinstance(first, str) else first.id
        return None

    def get_tasks(self):
        """ Returns the IDs of the tasks in the layout.
        """
        return [ (item if isinstance(item, str) else item.id)
                 for item in self.items ]

    def is_equivalent_to(self, layout):
        """ Returns whether two layouts are equivalent, i.e. whether they
            contain the same tasks.
        """
        return isinstance(layout, TaskWindowLayout) and \
            set(self.get_tasks()) == set(layout.get_tasks())
